/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller firmware routines                                      *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"

#include "firmwares.h"

static bool fw_get_builtin_firmware(struct firmware *fw, const char *name)
{
	int i;
	struct firmware_list *fw_list;
	extern struct firmware_list firmwares[];

	for ( i = 0; i < sizeof(firmwares) / sizeof(firmwares[0]); i++ ) {
		fw_list = &firmwares[i];
		if (strcmp(name, fw_list->name) == 0) {
			fw->data = (uint8_t *)fw_list->start;
			fw->size = fw_list->size;
			return true;
		}
	}

	return false;
}

static void release_firmware(struct firmware **firmware_p)
{
	if ( firmware_p != NULL && *firmware_p != NULL ) {
		free(*firmware_p, M_TEMP);
		*firmware_p = NULL;
	}
}

static int request_firmware(struct firmware **firmware_p, const char *name)
{
	struct firmware *firmware;

	init_firmwares();

	firmware =  malloc(sizeof(*firmware), M_TEMP, M_NOWAIT);
	if ( firmware == NULL ) {
		return ENOMEM;
	}
	memset(firmware, 0, sizeof(*firmware));

	if ( fw_get_builtin_firmware(firmware, name) ) {
		*firmware_p = firmware;
		return 0;
	}
	return ENOENT;
}

#define BUNDLESMALL 1
#define BUNDLEMAX (u16)6
#define INTDELAY (u16)1536 /* 0x600 */

/* Initialize firmware */
static const int e100_request_firmware(e100_dev_t *e100)
{
	const char *fw_name;
	struct firmware *fw = e100->fw;
	u8 timer, bundle, min_size;
	int err = 0;
	bool required = false;

	/* do not load u-code for ICH devices */
	if (e100->flags & ich)
		goto done;

	/* Search for ucode match against h/w revision
	 *
	 * Based on comments in the source code for the FreeBSD fxp
	 * driver, the FIRMWARE_D102E ucode includes both CPUSaver and
	 *
	 *    "fixes for bugs in the B-step hardware (specifically, bugs
	 *     with Inline Receive)."
	 *
	 * So we must fail if it cannot be loaded.
	 *
	 * The other microcode files are only required for the optional
	 * CPUSaver feature.  Nice to have, but no reason to fail.
	 */
	E100_DEBUG(log(LOG_ERR, "%s: E100 MAC %d", __devname__, e100->mac));
	if (e100->mac == mac_82559_D101M) {
		fw_name = FIRMWARE_D101M;
	} else if (e100->mac == mac_82559_D101S) {
		fw_name = FIRMWARE_D101S;
	} else if (e100->mac == mac_82551_F || e100->mac == mac_82551_10) {
		fw_name = FIRMWARE_D102E;
		required = true;
	} else { /* No ucode on other devices */
		goto done;
	}

	/* If the firmware has not previously been loaded, request a pointer
	 * to it. If it was previously loaded, we are reinitializing the
	 * adapter, possibly in a resume from hibernate, in which case
	 * request_firmware() cannot be used.
	 */
	if (!fw)
		err = request_firmware(&fw, fw_name);

	if (err) {
		if (required) {
			log(LOG_ERR, "%s: Failed to load firmware \"%s\": %d", __devname__,
				  fw_name, err);
			goto error_done;
		} else {
			log(LOG_WARNING, "%s: CPUSaver disabled. Needs \"%s\": %d", __devname__,
				  fw_name, err);
			goto done;
		}
	}

	/* Firmware should be precisely UCODE_SIZE (words) plus three bytes
	   indicating the offsets for BUNDLESMALL, BUNDLEMAX, INTDELAY */
	if (fw->size != (((UCODE_SIZE * 4 + 3) + 4U) & ~3U)) {
#ifdef __64__
		log(LOG_ERR, "%s: Firmware \"%s\" has wrong size %lu", __devname__,
#else
		log(LOG_ERR, "%s: Firmware \"%s\" has wrong size %u", __devname__,
#endif
			  fw_name, fw->size);
		release_firmware(&fw);
		errno = EINVAL;
		goto error_done;
	}

	/* Read timer, bundle and min_size from end of firmware blob */
	timer = fw->data[UCODE_SIZE * 4];
	bundle = fw->data[UCODE_SIZE * 4 + 1];
	min_size = fw->data[UCODE_SIZE * 4 + 2];

	if (timer >= UCODE_SIZE || bundle >= UCODE_SIZE ||
	    min_size >= UCODE_SIZE) {
		log(LOG_ERR, "%s: \"%s\" has bogus offset values (0x%x,0x%x,0x%x)", __devname__,
			  fw_name, timer, bundle, min_size);
		release_firmware(&fw);
		errno = EINVAL;
		goto error_done;
	}

	/* OK, firmware is validated and ready to use. Save a pointer
	 * to it in the e100 */
	e100->fw = fw;
done:
	return 0;
error_done:
	return -1;
}

static int e100_setup_ucode(e100_dev_t *e100, struct cb *cb)
{
	const struct firmware *fw = (struct firmware *)cb->data;
	u8 timer, bundle, min_size;

	/* It's not a real data; we just abused the fact that e100_exec_cb
	   will pass it through to here... */
	cb->data = NULL;

	/* firmware is stored as little endian already */
	memcpy(cb->u.ucode, fw->data, UCODE_SIZE * 4);

	/* Read timer, bundle and min_size from end of firmware blob */
	timer = fw->data[UCODE_SIZE * 4];
	bundle = fw->data[UCODE_SIZE * 4 + 1];
	min_size = fw->data[UCODE_SIZE * 4 + 2];

	/* Insert user-tunable settings in cb->u.ucode */
	cb->u.ucode[timer] &= cpu_to_le32(0xFFFF0000);
	cb->u.ucode[timer] |= cpu_to_le32(INTDELAY);
	cb->u.ucode[bundle] &= cpu_to_le32(0xFFFF0000);
	cb->u.ucode[bundle] |= cpu_to_le32(BUNDLEMAX);
	cb->u.ucode[min_size] &= cpu_to_le32(0xFFFF0000);
	cb->u.ucode[min_size] |= cpu_to_le32((BUNDLESMALL) ? 0xFFFF : 0xFF80);

	cb->command = cpu_to_le16(cb_ucode | cb_el);
	return 0;
}

int e100_load_ucode_wait(e100_dev_t *e100)
{
	int err = 0, counter = 50;
	struct cb *cb = e100->cb_to_clean;

	/* If it's NULL, then no ucode is required */
	if ( (err = e100_request_firmware(e100)) != 0 || e100->fw == NULL ) {
		if ( err )
			log(LOG_ERR, "%s: Request firmware is failed %d", __devname__, err);
		return err;
	}

	if ((err = e100_exec_cb(e100, (void *)e100->fw, e100_setup_ucode)))
		log(LOG_ERR, "%s: ucode cmd failed with error %d", __devname__, err);

	/* must restart cuc */
	e100->cuc_cmd = cuc_start;

	/* wait for completion */
	e100_write_flush(e100);
	udelay(10);

	/* wait for possibly (ouch) 500ms */
	while (!(cb->status & cpu_to_le16(cb_complete))) {
		msleep(10);
		if (!--counter) break;
	}

	/* ack any interrupts, something could have been set */
	iowrite8(~0, &e100->csr->scb.stat_ack);

	/* if the command failed, or is not OK, notify and return */
	if (!counter || !(cb->status & cpu_to_le16(cb_ok))) {
		log(LOG_ERR, "%s: ucode load failed", __devname__);
		err = -EPERM;
	}

	return err;
}
