/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller hw routines                                            *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"

void e100_write_flush(e100_dev_t *e100)
{
	/* Flush previous PCI writes through intermediate bridges
	 * by doing a benign read */
	(void)ioread8(&e100->csr->scb.status);
}

void e100_enable_irq(e100_dev_t *e100)
{
	unsigned irq_sw_gen_queue;
// 	InterruptLock(&e100->cmd_lock);
	if ( e100->stopping ) {
		e100_disable_irq(e100);
		return;
	}
	irq_sw_gen_queue = atomic_clr_value(&e100->irq_sw_gen, 0xFFFFFFFF);
	iowrite8(irq_sw_gen_queue | irq_mask_none, &e100->csr->scb.cmd_hi);
	e100_write_flush(e100);
// 	InterruptUnlock(&e100->cmd_lock);
}

void e100_disable_irq(e100_dev_t *e100)
{
// 	InterruptLock(&e100->cmd_lock);
	iowrite8(irq_mask_all, &e100->csr->scb.cmd_hi);
	e100_write_flush(e100);
// 	InterruptUnlock(&e100->cmd_lock);
}

void e100_hw_reset(e100_dev_t *e100)
{
	E100_DEBUG(
	pci_get_config(e100);
	)
	/* Put CU and RU into idle with a selective reset to get
	 * device off of PCI bus */
	iowrite32(selective_reset, &e100->csr->port);
	e100_write_flush(e100); udelay(20);

// 	/* Now fully reset device */
	iowrite32(software_reset, &e100->csr->port);
	e100_write_flush(e100); udelay(20);

	/* Mask off our interrupt line - it's unmasked after reset */
	e100_disable_irq(e100);
	E100_DEBUG(
	pci_get_config(e100);
	)
}

int e100_self_test(e100_dev_t *e100)
{
	u32 dma_addr = e100->dma_addr + offsetof(struct mem, selftest);

	/* Passing the self-test is a pretty good indication
	 * that the device can DMA to/from host memory */

	e100->mem->selftest.signature = 0;
	e100->mem->selftest.result = 0xFFFFFFFF;

	iowrite32(selftest | dma_addr, &e100->csr->port);
	e100_write_flush(e100);
	/* Wait 10 msec for self-test to complete */
	msleep(10);

	/* Interrupts are enabled after self-test */
	e100_disable_irq(e100);

	/* Check results of self-test */
	if (e100->mem->selftest.result != 0) {
		log(LOG_ERR, "%s: Self-test failed: result=0x%08X", __devname__,
			  e100->mem->selftest.result);
		return ETIMEDOUT;
	}
	if (e100->mem->selftest.signature == 0) {
		log(LOG_ERR, "%s: Self-test failed: timed out", __devname__);
		return ETIMEDOUT;
	}

	return 0;
}

#define E100_WAIT_SCB_TIMEOUT 20000 /* we might have to wait 100ms!!! */
#define E100_WAIT_SCB_FAST 20       /* delay like the old code */
int e100_exec_cmd(e100_dev_t *e100, u8 cmd, dma_addr_t dma_addr)
{
	unsigned int i;
	int err = 0;

// 	InterruptLock(&e100->cmd_lock);
	/* Previous command is accepted when SCB clears */
	for (i = 0; i < E100_WAIT_SCB_TIMEOUT; i++) {
		if (likely(!ioread8(&e100->csr->scb.cmd_lo)))
			break;
		if (unlikely(i > E100_WAIT_SCB_FAST))
			udelay(5);
	}
	if (unlikely(i == E100_WAIT_SCB_TIMEOUT)) {
		err = -EAGAIN;
		goto err_unlock;
	}

	if (unlikely(cmd != cuc_resume && cmd != ruc_resume && cmd != ruc_abort)) {
		if (dma_addr) {
// FIXME it's right?
			dma_addr += e100->bmstr;
		}
		iowrite32(dma_addr, &e100->csr->scb.gen_ptr);
	}
	iowrite8(cmd, &e100->csr->scb.cmd_lo);

err_unlock:
// 	InterruptUnlock(&e100->cmd_lock);
	return err;
}

int e100_exec_cb(e100_dev_t *e100, void *data,
	int (*cb_prepare)(e100_dev_t *, struct cb *))
{
// 	struct ifnet *ifp = &e100->ecom.ec_if;
	struct cb *cb;
	int err;

// 	InterruptLock(&e100->cb_lock);

	if (unlikely(!e100->cbs_avail)) {
		err = ENOMEM;
		goto err_unlock;
	}

	cb = e100->cb_to_use;
	e100->cb_to_use = cb->next;
	e100->cbs_avail--;
	cb->data = data;

	memset(&cb->u, 0, sizeof(cb->u));

	err = cb_prepare(e100, cb);
	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() cbs_avail %d command 0x%04x err %d", __devname__, __func__, e100->cbs_avail, cb->command, err);
	)
	if (err)
		goto err_unlock;

	if (unlikely(!e100->cbs_avail))
		err = ENOSPC;


	/* Order is important otherwise we'll be in a race with h/w:
	 * set S-bit in current first, then clear S-bit in previous. */
	cb->command |= cpu_to_le16(cb_s);
	dma_wmb();
	cb->prev->command &= cpu_to_le16(~cb_s);

	while (e100->cb_to_send != e100->cb_to_use) {
		if (unlikely(e100_exec_cmd(e100, e100->cuc_cmd,
			e100->cb_to_send->dma_addr))) {
			/* Ok, here's where things get sticky.  It's
			 * possible that we can't schedule the command
			 * because the controller is too busy, so
			 * let's just queue the command and try again
			 * when another command is scheduled. */
			E100_DEBUG(
				log(LOG_DEBUG, "%s: %s() EXEC CMD BUSY", __devname__, __func__);
			)
			break;
		} else {
			e100->cuc_cmd = cuc_resume;
			e100->cb_to_send = e100->cb_to_send->next;
		}
	}

err_unlock:
// 	InterruptUnlock(&e100->cb_lock);
	return err;
}

int e100_configure(e100_dev_t *e100, struct cb *cb)
{
	struct config *config = &cb->u.config;
// 	E100_DEBUG (
	u8 *c = (u8 *)config;
// 	)

	cb->command = cpu_to_le16(cb_config);

	memset(config, 0, sizeof(struct config));

	config->byte_count = 0x16;		/* bytes in this struct */
	config->rx_fifo_limit = 0x8;		/* bytes in FIFO before DMA */
	config->direct_rx_dma = 0x1;		/* reserved */
	config->standard_tcb = 0x1;		/* 1=standard, 0=extended */
	config->standard_stat_counter = 0x1;	/* 1=standard, 0=extended */
	config->rx_discard_short_frames = 0x1;	/* 1=discard, 0=pass */
	config->tx_underrun_retry = 0x3;	/* # of underrun retries */
	if (e100_phy_supports_mii(e100))
		config->mii_mode = 1;           /* 1=MII mode, 0=i82503 mode */
	config->pad10 = 0x6;
	config->no_source_addr_insertion = 0x1;	/* 1=no, 0=yes */
	config->preamble_length = 0x2;		/* 0=1, 1=3, 2=7, 3=15 bytes */
	config->ifs = 0x6;			/* x16 = inter frame spacing */
	config->ip_addr_hi = 0xF2;		/* ARP IP filter - not used */
	config->pad15_1 = 0x1;
	config->pad15_2 = 0x1;
	config->crs_or_cdt = 0x0;		/* 0=CRS only, 1=CRS or CDT */
	config->fc_delay_hi = 0x40;		/* time delay for fc frame */
	config->tx_padding = 0x1;		/* 1=pad short frames */
	config->fc_priority_threshold = 0x7;	/* 7=priority fc disabled */
	config->pad18 = 0x1;
	config->full_duplex_pin = 0x1;		/* 1=examine FDX# pin */
	config->pad20_1 = 0x1F;
	config->fc_priority_location = 0x1;	/* 1=byte#31, 0=byte#19 */
	config->pad21_1 = 0x5;

	config->adaptive_ifs = e100->adaptive_ifs;
	config->loopback = e100->loopback;

	if (e100->force_advertise && e100->cfg.duplex > 0)
		config->full_duplex_force = 0x1;	/* 1=force, 0=auto */

	if (e100->flags & promiscuous || e100->loopback) {
		config->rx_save_bad_frames = 0x1;	/* 1=save, 0=discard */
		config->rx_discard_short_frames = 0x0;	/* 1=discard, 0=save */
		config->promiscuous_mode = 0x1;		/* 1=on, 0=off */
	}


	if (e100->flags & multicast_all)
		config->multicast_all = 0x1;		/* 1=accept, 0=no */

	/* disable WoL when up */
	if (netif_running(e100) || !(e100->flags & wol_magic))
		config->magic_packet_disable = 0x1;	/* 1=off, 0=on */

	if (e100->mac >= mac_82558_D101_A4) {
		config->fc_disable = 0x1;	/* 1=Tx fc off, 0=Tx fc on */
		if ( (e100->pci_command & PCI_COMMAND_INVALIDATE_ENABLE) )
			config->mwi_enable = 0x1;	/* 1=enable, 0=disable */
		config->rx_long_ok = 0x1;	/* 1=VLANs ok, 0=standard */
		if (e100->mac >= mac_82559_D101M) {
			config->tno_intr = 0x1;		/* TCO stats enable */
			/* Enable TCO in extended config */
#if 1
// FIXME
			if (e100->mac >= mac_82551_10) {
				config->byte_count = 0x20; /* extended bytes */
				config->rx_d102_mode = 0x1; /* GMRC for TCO */
			}
#endif
		} else {
			config->standard_stat_counter = 0x0;
		}
	}

// 	E100_DEBUG (
	if ( e100->cfg.verbose > 1 ) {
		log(LOG_DEBUG, "%s: config [00-07]=%08xh", __devname__, *(u32 *)c); c += 8;
		log(LOG_DEBUG, "%s: config [08-15]=%08xh", __devname__, *(u32 *)c); c += 8;
		log(LOG_DEBUG, "%s: config [16-23]=%08xh", __devname__, *(u32 *)c);
	}
// 	)
	return 0;
}

int e100_setup_iaaddr(e100_dev_t *e100, struct cb *cb)
{
	cb->command = cpu_to_le16(cb_iaaddr);
	memcpy(cb->u.iaaddr, e100->cfg.current_address, ETH_MAC_LEN);
	return 0;
}

static int e100_multi(e100_dev_t *e100, struct cb *cb)
{
	struct ethercom			*ec;
	struct ether_multi		*enm;
	struct ether_multistep	step;
	int						mcentries = 0;

	ec = &(e100->ecom);

	cb->command = cpu_to_le16(cb_multi);
	ETHER_FIRST_MULTI (step, ec, enm);
	while (enm != NULL) {
		if ((mcentries > E100_MAX_MULTICAST_ADDRS) || (memcmp (enm->enm_addrlo, enm->enm_addrhi,
			ETHER_ADDR_LEN)) != 0) {
			ec->ec_if.if_flags |= IFF_ALLMULTI;
			break;
		}
		if ( e100->cfg.verbose > 1 ) {
			log(LOG_DEBUG,"%s: enm %p %02X:%02X:%02X mcentries %d",  __devname__, enm, 
				enm->enm_addrlo[0], enm->enm_addrlo[1], enm->enm_addrlo[2], mcentries);
		}
		memcpy(&cb->u.multi.addr[mcentries * ETH_MAC_LEN], &enm->enm_addrlo, ETH_MAC_LEN);
		mcentries++;
		ETHER_NEXT_MULTI (step, enm);
	}
	cb->u.multi.count = cpu_to_le16(min(mcentries, E100_MAX_MULTICAST_ADDRS) * ETH_MAC_LEN);
	return 0;
}

static int e100_mc_count(e100_dev_t *e100)
{
	struct ethercom			*ec;
	struct ether_multi		*enm;
	struct ether_multistep	step;
	int						mcentries = 0;

	ec = &e100->ecom;
	ETHER_FIRST_MULTI (step, ec, enm);
	while (enm != NULL) {
		if ((mcentries > E100_MAX_MULTICAST_ADDRS) || (memcmp (enm->enm_addrlo, enm->enm_addrhi,
			ETHER_ADDR_LEN)) != 0) {
			ec->ec_if.if_flags |= IFF_ALLMULTI;
			mcentries = E100_MAX_MULTICAST_ADDRS + 1;
			return mcentries;
		}
		if ( e100->cfg.verbose > 1 ) {
			log(LOG_DEBUG,"%s: enm %p %02X:%02X:%02X mcentries %d",  __devname__, enm, 
				enm->enm_addrlo[0], enm->enm_addrlo[1], enm->enm_addrlo[2], mcentries);
		}
		mcentries++;
		ETHER_NEXT_MULTI (step, enm);
	}
	return mcentries;
}

void e100_set_multicast_list(e100_dev_t *e100)
{
	struct ifnet			*ifp;
	int mc_count = 0;

	ifp = &e100->ecom.ec_if;

	if ( (e100->cfg.flags & NIC_FLAG_MULTICAST) ) {
		ifp->if_flags &= ~IFF_ALLMULTI;
		if (ifp->if_flags & IFF_PROMISC) {
			e100->cfg.flags |= NIC_FLAG_PROMISCUOUS;
			e100->flags |= promiscuous;
		} else {
			e100->cfg.flags &= ~NIC_FLAG_PROMISCUOUS;
			e100->flags &= ~promiscuous;
		}

		if ( (mc_count = e100_mc_count(e100)) > E100_MAX_MULTICAST_ADDRS)
			e100->flags |= multicast_all;
		else
			e100->flags &= ~multicast_all;
	} else {
		e100->flags &= ~multicast_all;
	}

	e100_exec_cb(e100, NULL, e100_configure);
	if ( (e100->cfg.flags & NIC_FLAG_MULTICAST) ) {
		e100_exec_cb(e100, NULL, e100_multi);
	}
}

int e100_hw_init(e100_dev_t *e100)
{
	int err = 0;

	e100_hw_reset(e100);

	E100_DEBUG(log(LOG_DEBUG, "%s: e100_hw_init", __devname__));

	if ((err = e100_self_test(e100))) {
		log(LOG_DEBUG, "%s: Self test is failed: %d", __devname__, err);
		return err;
	}

	if ((err = e100_findphy(e100))) {
		log(LOG_ERR, "%s: Find PHY is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_exec_cmd(e100, cuc_load_base, 0))) {
		log(LOG_ERR, "%s: Execute of 'cuc load base' is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_exec_cmd(e100, ruc_load_base, 0))) {
		log(LOG_ERR, "%s: Execute of 'ruc load base' is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_load_ucode_wait(e100))) {
		log(LOG_ERR, "%s: Load ucode is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_exec_cb(e100, NULL, e100_configure))) {
		log(LOG_ERR, "%s: Execute of 'configure' is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_exec_cb(e100, NULL, e100_setup_iaaddr))) {
		log(LOG_ERR, "%s: Execute of 'setup iaddr' is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_exec_cmd(e100, cuc_dump_addr,
		e100->dma_addr + offsetof(struct mem, stats)))) {
		log(LOG_ERR, "%s: Execute of 'setup cuc dump addr' is failed: %d", __devname__, err);
		return err;
	}
	if ((err = e100_exec_cmd(e100, cuc_dump_reset, 0))) {
		log(LOG_ERR, "%s: Execute of 'setup cuc dump reset' is failed: %d", __devname__, err);
		return err;
	}

	e100_disable_irq(e100);

	return 0;
}

#if 0
static void e100_eeprom_write(e100_dev_t *e100, u16 addr_len, u16 addr, __le16 data)
{
	u32 cmd_addr_data[3];
	u8 ctrl;
	int i, j;

	/* Three cmds: write/erase enable, write data, write/erase disable */
	cmd_addr_data[0] = op_ewen << (addr_len - 2);
	cmd_addr_data[1] = (((op_write << addr_len) | addr) << 16) |
		le16_to_cpu(data);
	cmd_addr_data[2] = op_ewds << (addr_len - 2);

	/* Bit-bang cmds to write word to eeprom */
	for (j = 0; j < 3; j++) {

		/* Chip select */
		iowrite8(eecs | eesk, &e100->csr->eeprom_ctrl_lo);
		e100_write_flush(e100); udelay(4);

		for (i = 31; i >= 0; i--) {
			ctrl = (cmd_addr_data[j] & (1 << i)) ?
				eecs | eedi : eecs;
			iowrite8(ctrl, &e100->csr->eeprom_ctrl_lo);
			e100_write_flush(e100); udelay(4);

			iowrite8(ctrl | eesk, &e100->csr->eeprom_ctrl_lo);
			e100_write_flush(e100); udelay(4);
		}
		/* Wait 10 msec for cmd to complete */
		msleep(10);

		/* Chip deselect */
		iowrite8(0, &e100->csr->eeprom_ctrl_lo);
		e100_write_flush(e100); udelay(4);
	}
};

/* Save (portion of) driver EEPROM cache to device and update checksum */
static int e100_eeprom_save(e100_dev_t *e100, u16 start, u16 count)
{
	u16 addr, addr_len = 8, checksum = 0;

	/* Try reading with an 8-bit addr len to discover actual addr len */
	e100_eeprom_read(e100, &addr_len, 0);
	e100->eeprom_wc = 1 << addr_len;

	if (start + count >= e100->eeprom_wc)
		return -EINVAL;

	for (addr = start; addr < start + count; addr++)
		e100_eeprom_write(e100, addr_len, addr, e100->eeprom[addr]);

	/* The checksum, stored in the last word, is calculated such that
	 * the sum of words should be 0xBABA */
	for (addr = 0; addr < e100->eeprom_wc - 1; addr++)
		checksum += le16_to_cpu(e100->eeprom[addr]);
	e100->eeprom[e100->eeprom_wc - 1] = cpu_to_le16(0xBABA - checksum);
	e100_eeprom_write(e100, addr_len, e100->eeprom_wc - 1,
		e100->eeprom[e100->eeprom_wc - 1]);

	return 0;
}
#endif
/* General technique stolen from the eepro100 driver - very clever */
static __le16 e100_eeprom_read(e100_dev_t *e100, u16 *addr_len, u16 addr)
{
	u32 cmd_addr_data;
	u16 data = 0;
	u8 ctrl;
	int i;

	cmd_addr_data = ((op_read << *addr_len) | addr) << 16;

	/* Chip select */
	iowrite8(eecs | eesk, &e100->csr->eeprom_ctrl_lo);
	e100_write_flush(e100); udelay(4);

	/* Bit-bang to read word from eeprom */
	for (i = 31; i >= 0; i--) {
		ctrl = (cmd_addr_data & (1 << i)) ? eecs | eedi : eecs;
		iowrite8(ctrl, &e100->csr->eeprom_ctrl_lo);
		e100_write_flush(e100); udelay(4);

		iowrite8(ctrl | eesk, &e100->csr->eeprom_ctrl_lo);
		e100_write_flush(e100); udelay(4);

		/* Eeprom drives a dummy zero to EEDO after receiving
		 * complete address.  Use this to adjust addr_len. */
		ctrl = ioread8(&e100->csr->eeprom_ctrl_lo);
		if (!(ctrl & eedo) && i > 16) {
			*addr_len -= (i - 16);
			i = 17;
		}

		data = (data << 1) | (ctrl & eedo ? 1 : 0);
	}

	/* Chip deselect */
	iowrite8(0, &e100->csr->eeprom_ctrl_lo);
	e100_write_flush(e100); udelay(4);

	return cpu_to_le16(data);
};

/* Load entire EEPROM image into driver cache and validate checksum */
int e100_eeprom_load(e100_dev_t *e100)
{
	u16 addr, addr_len = 8, checksum = 0;

	/* Try reading with an 8-bit addr len to discover actual addr len */
	e100_eeprom_read(e100, &addr_len, 0);
	e100->eeprom_wc = 1 << addr_len;

	for (addr = 0; addr < e100->eeprom_wc; addr++) {
		e100->eeprom[addr] = e100_eeprom_read(e100, &addr_len, addr);
		if (addr < e100->eeprom_wc - 1) {
			checksum += le16_to_cpu(e100->eeprom[addr]);
			E100_DEBUG(
			log(LOG_ERR, "%s: checksum += eeprom[%d] %04X = %04X", __devname__, addr, e100->eeprom[addr], checksum);
			)
		}
	}
	E100_DEBUG(
	log(LOG_ERR, "%s: eeprom[%d] %04X", __devname__, e100->eeprom_wc - 1, e100->eeprom[e100->eeprom_wc - 1]);
	)
	/* The checksum, stored in the last word, is calculated such that
	 * the sum of words should be 0xBABA */
	if (cpu_to_le16(0xBABA - checksum) != e100->eeprom[e100->eeprom_wc - 1]) {
		log(LOG_ERR, "%s: EEPROM corrupted (%d)", __devname__, addr_len);
		if (!e100->eeprom_bad_csum_allow)
			return EAGAIN;
	}

	return 0;
}

int e100_alloc_cbs(e100_dev_t *e100)
{
	struct cb *cb;
	unsigned int i, count = e100->params.cbs.count;

	e100->cuc_cmd = cuc_start;
	e100->cb_to_use = e100->cb_to_send = e100->cb_to_clean = NULL;
	e100->cbs_avail = 0;

	e100->cbs = (struct cb *)mmap(0, count * sizeof(struct cb), e100->prot_flags, e100->map_flags, e100->tmem_fd, 0);
	if (e100->cbs == (struct cb *)MAP_FAILED) {
		return errno;
	}
	e100->cbs_dma_addr = (dma_addr_t)drvr_mphys(e100->cbs);

	memset(e100->cbs, 0, count * sizeof(struct cb));
	for (cb = e100->cbs, i = 0; i < count; cb++, i++) {
		cb->next = (i + 1 < count) ? cb + 1 : e100->cbs;
		cb->prev = (i == 0) ? e100->cbs + count - 1 : cb - 1;

		cb->dma_addr = e100->cbs_dma_addr + i * sizeof(struct cb);
		cb->link = cpu_to_le32(e100->cbs_dma_addr +
			((i+1) % count) * sizeof(struct cb) + e100->bmstr);
	}

	e100->cb_to_use = e100->cb_to_send = e100->cb_to_clean = e100->cbs;
	e100->cbs_avail = count;

	E100_DEBUG(log(LOG_DEBUG, "%s: ALLOC CBS %d", __devname__, count));

	return 0;
}

void e100_clean_cbs(e100_dev_t *e100, int full_clean)
{
	if (e100->cbs) {
		while ( e100_tx_clean(e100) )
			;
		if ( full_clean ) {
			munmap(e100->cbs, e100->params.cbs.count * sizeof(struct cb));
			e100->cbs = NULL;
			e100->cbs_avail = 0;
		}
	}
	e100->cuc_cmd = cuc_start;
	e100->cb_to_use = e100->cb_to_send = e100->cb_to_clean = e100->cbs;
	E100_DEBUG(log(LOG_DEBUG, "%s: Clean CBS %d -> %d", __devname__, e100->params.cbs.count, e100->cbs_avail));
}

void e100_hw_dump_registers(e100_dev_t *e100, int desc)
{
	log(LOG_ERR, "%s (%s) Logical Net %d, Reg : 0x%08x",
				e100->cfg.device_description, __devname__, e100->cfg.lan, FROM_POINTER_TO_INT(e100->csr, uint32_t));

	if ( e100->mdi != NULL )
		MDI_DumpRegisters( e100->mdi, e100->cfg.phy_addr );
}
