/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller misc routines                                          *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"

int netif_running(e100_dev_t *e100)
{
	struct ifnet *ifp;
	ifp = &e100->ecom.ec_if;
	return (ifp->if_flags & IFF_RUNNING);
}

int early_transmit(e100_dev_t *e100, struct nw_work_thread *wtp)
{
	struct ifnet *ifp = &e100->ecom.ec_if;

	if ( (ifp->if_flags_tx & IFF_RUNNING) == 0 ||
			(e100->cfg.flags & NIC_FLAG_LINK_DOWN) ) {
		return 2;
	}
	if ( e100->start_running ) {
		return 1;
	}
	return 0;
}

// Get system memory cache attrs
static int mem_smart_cache(void)
{
#if 1
	struct cacheattr_entry	*cache, *base;
	int						i;

	base = SYSPAGE_ENTRY(cacheattr);
	for (i = SYSPAGE_ENTRY(cpuinfo)->data_cache; i != CACHE_LIST_END; i = cache->next) {
		cache = &base[i];
		if (!(cache->flags & CACHE_FLAG_SNOOPED))
			return 0;
	}
	return 1;
#else
	return 0;
#endif
}

void setup_mmap_flags(e100_dev_t *e100)
{
	/* Prepare mmap flags */
	if (e100->tmem_fd != NOFD)
		e100->map_flags = MAP_SHARED;
	else
		e100->map_flags = MAP_PRIVATE | MAP_PHYS | MAP_ANON;
	
	if (mem_smart_cache())
		e100->prot_flags = PROT_READ | PROT_WRITE;
	else
		e100->prot_flags = PROT_READ | PROT_WRITE | PROT_NOCACHE;

	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() PROT_NOCACHE %d", __devname__, __func__, e100->prot_flags & PROT_NOCACHE);
	)
}

void setup_tmem(e100_dev_t *e100)
{
	if ( e100->tmem == NULL ) {
		goto notmem;
	}
	if ((e100->tmem_fd = posix_typed_mem_open(e100->tmem, O_RDWR,
		POSIX_TYPED_MEM_ALLOCATE_CONTIG)) == -1) {
		log(LOG_ERR, "%s: unable to open typed memory %s, will not use typed memory", 
													__devname__, e100->tmem);
		goto notmem;
	}
	return;
notmem:
	e100->tmem_fd = NOFD;
}

int pci_get_config(e100_dev_t *e100)
{
	if ( pci_read_config(e100->pci_dev_hdl, offsetof (struct _pci_config_regs, Command), 1, 2, &e100->pci_command) == -1 ) {
		return ENODEV;
	}
	if ( pci_read_config(e100->pci_dev_hdl, offsetof (struct _pci_config_regs, Status), 1, 2, &e100->pci_status) == -1 ) {
		return ENODEV;
	}
	E100_DEBUG(
	log(LOG_INFO, "%s: Command %04X Status %04X", __devname__, e100->pci_command, e100->pci_status);
	)

	return 0;
}

int pci_set_master(e100_dev_t *e100)
{
	if ( e100->pci_dev_hdl == NULL ) {
		return ENOMEM;
	}
	pci_attach_device(e100->pci_dev_hdl,
	    PCI_MASTER_ENABLE, e100->cfg.device_index, &e100->pci_info);

	return pci_get_config(e100);
}

int pci_enable_device(e100_dev_t *e100)
{
	e100->pci_dev_hdl = pci_attach_device(NULL, PCI_INIT_ALL | PCI_MASTER_ENABLE, e100->cfg.device_index, &e100->pci_info);
	e100->cfg.device_revision = e100->pci_info.Revision;
	if (e100->pci_dev_hdl == NULL) {
		return errno;
	}

	return pci_get_config(e100);
}

void pci_disable_device(e100_dev_t *e100)
{
	int ret;
	ret = pci_detach_device(e100->pci_dev_hdl);
	E100_DEBUG(log(LOG_ERR, "%s: PCI disable device %d", __devname__, ret));
	(void)ret;
}

void fill_random_mac(char *mac_addr)
{
	int i;
	uint64_t	clc = ClockCycles();
	uint32_t	rnd = ((clc >> 32) & 0xFFFFFF) ^ (clc & 0xFFFFFF);
	for ( i = 0; i < 6; i++ ) {
		rnd = (0xFE0303D9 * rnd + 0x2F77DF2F);
		mac_addr[i] = rnd >> 24;
	}
	mac_addr[0] &= ~1;
	mac_addr[0] |= 2;
}

int is_broadcast_mac(uint8_t *mac_addr)
{
	return !memcmp(mac_addr, "\xFF\xFF\xFF\xFF\xFF\xFF", ETH_MAC_LEN);
}

int valid_mac_addr(char *mac_addr)
{
	return (mac_addr[0] & 1 || !memcmp(mac_addr, "\0\0\0\0\0\0", ETH_MAC_LEN)) ? 0 : 1;
}

int e100_set_ringparam(e100_dev_t *e100, int num_receive, int num_transmit, int tx_reap)
{
	struct	ifnet *ifp;
	struct param_range *rfds = &e100->params.rfds;
	struct param_range *cbs = &e100->params.cbs;
	int e100_running = netif_running(e100);

	ifp = &e100->ecom.ec_if;

	if (e100_running)
		e100_stop(ifp, 0);

	if ( num_receive >= 0 ) {
		rfds->count = max(num_receive, rfds->min);
		rfds->count = min(rfds->count, rfds->max);
	}
	if ( num_transmit >= 0 ) {
		cbs->count = max(num_transmit, cbs->min);
		cbs->count = min(cbs->count, cbs->max);
	}
	if ( tx_reap >= 0 ) {
		cbs->reap = max(tx_reap, cbs->min);
		cbs->reap = min(cbs->reap, cbs->max);
	}
	if ( e100->cfg.verbose && (num_receive >= 0 || num_transmit >= 0 || tx_reap >= 0) )
		log(LOG_INFO, "%s: Ring Param settings: rx: %d, tx %d, reap %d", __devname__, rfds->count, cbs->count, cbs->reap);
	if (e100_running)
		e100_init(ifp);

	return 0;
}

void e100_get_defaults(e100_dev_t *e100)
{
	struct param_range rfds = { .min = 16, .max = 2048, .count = 256, .reap = 0 };
	struct param_range cbs  = { .min = 64, .max = 2048, .count = 1024, .reap = 64 };

	/* MAC type is encoded as rev ID; exception: ICH is treated as 82559 */
	e100->mac = (e100->flags & ich) ? mac_82559_D101M : e100->pci_info.Revision;
	if (e100->mac == mac_unknown)
		e100->mac = mac_82557_D100_A;

	e100->params.rfds = rfds;
	e100->params.cbs = cbs;

	e100_set_ringparam(e100, rfds.count, cbs.count, cbs.reap);

	/* Quadwords to DMA into FIFO before starting frame transmit */
	e100->tx_threshold = 0xE0;

	/* no interrupt for every tx completion, delay = 256us if not 557 */
	e100->tx_command = cpu_to_le16(cb_tx | cb_tx_sf |
		((e100->mac >= mac_82558_D101_A4) ? cb_cid : cb_i));

	/* Template for a freshly allocated RFD */
	e100->blank_rfd.command = 0;
	e100->blank_rfd.rbd = cpu_to_le32(0xFFFFFFFF);
	e100->blank_rfd.size = cpu_to_le16(MAX_ETHER_LEN);

	if ( e100->num_receive >=0 || e100->num_transmit >= 0 || e100->tx_reap ) {
		e100_set_ringparam(e100, e100->num_receive, e100->num_transmit, e100->tx_reap);
	}
}

void e100_adjust_adaptive_ifs(e100_dev_t *e100, int speed, int duplex)
{
	/* Adjust inter-frame-spacing (IFS) between two transmits if
	 * we're getting collisions on a half-duplex connection. */

	if (duplex == 0) {
		u32 prev = e100->adaptive_ifs;
		u32 min_frames = (speed == 100) ? 1000 : 100;

		if ((e100->tx_frames / 32 < e100->tx_collisions) &&
		   (e100->tx_frames > min_frames)) {
			if (e100->adaptive_ifs < 60)
				e100->adaptive_ifs += 5;
		} else if (e100->tx_frames < min_frames) {
			if (e100->adaptive_ifs >= 5)
				e100->adaptive_ifs -= 5;
		}
		if (e100->adaptive_ifs != prev)
			e100_exec_cb(e100, NULL, e100_configure);
	}

	if (e100->flags & ich && speed == 10 && duplex == 0)
		/* Need SW workaround for ICH[x] 10Mbps/half duplex Tx hang. */
		e100->flags |= ich_10h_workaround;
	else
		e100->flags &= ~ich_10h_workaround;
}

void e100_update_stats(e100_dev_t *e100)
{
	nic_stats_t *stats = &e100->stats;
	nic_ethernet_stats_t *estats = &stats->un.estats;
	struct stats *s = &e100->mem->stats;
	__le32 *complete = (e100->mac < mac_82558_D101_A4) ? &s->fc_xmt_pause :
		(e100->mac < mac_82559_D101M) ? (__le32 *)&s->xmt_tco_frames :
		&s->complete;

	/* Device's stats reporting may take several microseconds to
	 * complete, so we're always waiting for results of the
	 * previous command. */

	if (*complete == cpu_to_le32(cuc_dump_reset_complete)) {
		*complete = 0;

		e100->tx_frames = le32_to_cpu(s->tx_good_frames);
		e100->tx_collisions = le32_to_cpu(s->tx_total_collisions);
		estats->xcoll_aborted += le32_to_cpu(s->tx_max_collisions);
		estats->late_collisions += le32_to_cpu(s->tx_late_collisions);
		estats->tx_deferred += le32_to_cpu(s->tx_deferred);
		estats->single_collisions += le32_to_cpu(s->tx_single_collisions);
		estats->multi_collisions += le32_to_cpu(s->tx_multiple_collisions);
		estats->no_carrier += le32_to_cpu(s->tx_lost_crs);
		estats->internal_tx_errors += le32_to_cpu(s->tx_underruns);
		estats->total_collision_frames += e100->tx_collisions;
		estats->short_packets += le32_to_cpu(s->rx_short_frame_errors);
		estats->oversized_packets = e100->rx_over_length_errors;
		estats->fcs_errors += le32_to_cpu(s->rx_crc_errors);
		estats->align_errors += le32_to_cpu(s->rx_alignment_errors);

		estats->internal_rx_errors += le32_to_cpu(s->rx_overrun_errors);
		estats->internal_rx_errors += le32_to_cpu(s->rx_resource_errors);
		estats->internal_rx_errors += le32_to_cpu(s->rx_cdt_errors);

		if (e100->mac >= mac_82558_D101_A4) {
			e100->tx_fc_pause += le32_to_cpu(s->fc_xmt_pause);
			e100->rx_fc_pause += le32_to_cpu(s->fc_rcv_pause);
			e100->rx_fc_unsupported +=
				le32_to_cpu(s->fc_rcv_unsupported);
			if (e100->mac >= mac_82559_D101M) {
				e100->tx_tco_frames +=
					le16_to_cpu(s->xmt_tco_frames);
				e100->rx_tco_frames +=
					le16_to_cpu(s->rcv_tco_frames);
			}
		}
	}


	if (e100_exec_cmd(e100, cuc_dump_reset, 0))
		log(LOG_ERR, "%s: exec cuc_dump_reset failed", __devname__);
}
