/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller main routines                                          *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"

static int e100_attach(struct device *parent, struct device *self, void *aux);
static int e100_detach(struct device *dev, int flags);
static void e100_shutdown(void *arg);
static void e100_destroy(e100_dev_t *e100, enum destroy_variants dstage);
static int e100_alloc(e100_dev_t *e100);
static void e100_free(e100_dev_t *e100);
static int e100_inter_entry_setup(e100_dev_t *e100);
static void ifp_setup(e100_dev_t *e100, const char *ifname);
static int e100_parse_options(e100_dev_t *e100, const char *optstring, unsigned instance, nic_config_t *cfg);
static int e100_entry(void *dll_hdl,  struct _iopkt_self *iopkt, char *options);
static void e100_hk_callout(void *arg);
int e100_init(struct ifnet *ifp);
void e100_stop(struct ifnet *ifp, int disable);

struct _iopkt_drvr_entry IOPKT_DRVR_ENTRY_SYM(e100) = IOPKT_DRVR_ENTRY_SYM_INIT(e100_entry);

CFATTACH_DECL(e100,
	sizeof(struct e100_dev),
	NULL,
	e100_attach,
	e100_detach,
	NULL);

struct e100_attach_args {
	void		*dll_hdl;
	char 		*options;
	unsigned	instance;
	unsigned	vendor_id;
	unsigned	device_id;
	unsigned	bus;
	unsigned	devfn;
	int			device_idx;
	unsigned	driver_data;
};

#define INTEL_8255X_ETHERNET_DEVICE(device_id, ich) {\
	PCI_VENDOR_ID_INTEL, device_id, ich }
static struct pci_device_id known_device_ids[] = {
	INTEL_8255X_ETHERNET_DEVICE(0x1029, 0),
	INTEL_8255X_ETHERNET_DEVICE(0x1030, 0),
	INTEL_8255X_ETHERNET_DEVICE(0x1031, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1032, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1033, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1034, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1035, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1036, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1037, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1038, 3),
	INTEL_8255X_ETHERNET_DEVICE(0x1039, 4),
	INTEL_8255X_ETHERNET_DEVICE(0x103A, 4),
	INTEL_8255X_ETHERNET_DEVICE(0x103B, 4),
	INTEL_8255X_ETHERNET_DEVICE(0x103C, 4),
	INTEL_8255X_ETHERNET_DEVICE(0x103D, 4),
	INTEL_8255X_ETHERNET_DEVICE(0x103E, 4),
	INTEL_8255X_ETHERNET_DEVICE(0x1050, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1051, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1052, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1053, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1054, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1055, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1056, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1057, 5),
	INTEL_8255X_ETHERNET_DEVICE(0x1059, 0),
	INTEL_8255X_ETHERNET_DEVICE(0x1064, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x1065, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x1066, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x1067, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x1068, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x1069, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x106A, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x106B, 6),
	INTEL_8255X_ETHERNET_DEVICE(0x1091, 7),
	INTEL_8255X_ETHERNET_DEVICE(0x1092, 7),
	INTEL_8255X_ETHERNET_DEVICE(0x1093, 7),
	INTEL_8255X_ETHERNET_DEVICE(0x1094, 7),
	INTEL_8255X_ETHERNET_DEVICE(0x1095, 7),
	INTEL_8255X_ETHERNET_DEVICE(0x10fe, 7),
	INTEL_8255X_ETHERNET_DEVICE(0x1209, 0),
	INTEL_8255X_ETHERNET_DEVICE(0x1229, 0),
	INTEL_8255X_ETHERNET_DEVICE(0x2449, 2),
	INTEL_8255X_ETHERNET_DEVICE(0x2459, 2),
	INTEL_8255X_ETHERNET_DEVICE(0x245D, 2),
	INTEL_8255X_ETHERNET_DEVICE(0x27DC, 7),
	{ 0, }
};

static int e100_entry(void *dll_hdl,  struct _iopkt_self *iopkt, char *options)
{
	nic_config_t			cfg;
	int				devid, idx;
	unsigned			bus, dev_func;
	int				rc;
	int				instance, status, single;
	struct e100_attach_args	e100_args;
	struct device			*dev;

	single = 0;
	memset(&cfg, 0, sizeof(cfg));

	
	if (options != NULL) {
		cfg.vendor_id = cfg.device_id = cfg.device_index = 0xffffffff;
		if ((rc = e100_parse_options(NULL, options, -1, &cfg)) != 0)
			return rc;
		if (cfg.device_id != 0xffffffff) {
			if (cfg.vendor_id == 0xffffffff) {
				cfg.vendor_id = PCI_VENDOR_ID_INTEL;
			}
			known_device_ids[0].device = cfg.device_id;
			known_device_ids[0].vendor = cfg.vendor_id;
			known_device_ids[1].vendor = known_device_ids[1].device = 0;
		}
	} else {
		cfg.device_index = 0xffffffff;
	}

	if (nw_pci_hdl == -1) {
		log(LOG_ERR, "%s: Could not attach to PCI server", "fxp*");
		return EAGAIN;
	}

	rc = ENODEV;

	instance = 0;

	for (devid = 0; known_device_ids[devid].vendor != 0; devid++) {
		idx = ((cfg.device_index == 0xffffffff) ? 0 : cfg.device_index);
		while (1) {
			E100_DEBUG2(
			log(LOG_DEBUG, "fxp%d: PCI find device %04X:%04X.%d", instance, known_device_ids[devid].vendor, known_device_ids[devid].device, idx)
			);
			if ((status = pci_find_device(known_device_ids[devid].device,
			    known_device_ids[devid].vendor, idx, &bus, &dev_func)) !=
			    PCI_SUCCESS)
				break;

			e100_args.dll_hdl = dll_hdl;
			e100_args.instance = instance;
			e100_args.options = options;
			e100_args.vendor_id = known_device_ids[devid].vendor;
			e100_args.device_id = known_device_ids[devid].device;
			e100_args.bus = bus;
			e100_args.devfn = dev_func;
			e100_args.device_idx = idx;
			e100_args.driver_data = known_device_ids[devid].driver_data;

			dev = NULL;
			if ((rc = dev_attach("fxp", options, &e100_ca,
			    &e100_args, &single, &dev, NULL)) != EOK) {
				goto done;
			}
			instance++;
			if (cfg.device_index != 0xffffffff || single != 0)
				goto done;
			idx++;
		}
	}

done:
	if (instance)
		return EOK;
	return rc;
}

enum e100_opts_enum {
	E100_RECEIVE,
	E100_TRANSMIT,
	E100_TX_REAP,
	E100_USE_IO,
	E100_KERMASK,
	E100_EEPROM_BAD_CSUM_ALLOW,
	E100_FORCE_RX_BUG,
	E100_EXTRA_RNR,
	E100_HK_TIMEOUT,
	TYPED_MEM
};

static char *e100_opts[] = {
	"receive",
	"transmit",
	"tx_reap",
	"use_io",
	"kermask",
	"eeprom_bad_csum_allow",
	"force_rx_bug",
	"extra_rnr",
	"hk_timeout",
	"typed_mem",
	NULL
};

static int e100_parse_options(e100_dev_t *e100, const char *optstring, unsigned instance, nic_config_t *cfg)
{
	char		*value;
	int		opt;
	char		*options, *freeptr;
	char		*c;
	int		rc = 0;
	

	if ( e100 != NULL ) {
		e100->num_transmit = -1;
		e100->num_receive = -1;
		e100->tx_reap = -1;
		e100->hk_timeout = E100_DEFAULT_HK_TIMEOUT;
#ifdef __E2K__
		e100->tmem = "/memory/below4G/ram";
#endif
	}
	if (optstring == NULL)
		return 0;

	/* getsubopt() is destructive */
	options = malloc(strlen(optstring) + 1, M_TEMP, M_NOWAIT);
	if (options == NULL)
		return ENOMEM;
	strcpy(options, optstring);
	freeptr = options;

	while (options && *options != '\0') {
		c = options;
		if ((opt = getsubopt(&options, e100_opts, &value)) == -1) {
			if (nic_parse_options(cfg, value) == EOK)
				continue;
			goto error;
		}

		if (e100 == NULL)
			continue;

		switch (opt) {
		case E100_RECEIVE:
			e100->num_receive = strtoul(value, 0, 0);
			continue;

		case E100_TRANSMIT:
			e100->num_transmit = strtoul(value, 0, 0);
			continue;

		case E100_TX_REAP:
			e100->tx_reap = strtoul(value, 0, 0);
			continue;

		case E100_USE_IO:
			e100->use_io = 1;
			continue;

		case E100_KERMASK:
			if (value != NULL) {
				e100->kermask = strtol(value, 0, 0);
				if (e100->kermask != 0)
					e100->kermask = 1;
			}
			continue;

		case E100_EEPROM_BAD_CSUM_ALLOW:
			if (value != NULL) {
				e100->eeprom_bad_csum_allow = strtoul(value, 0, 0);
			}
			continue;

		case E100_FORCE_RX_BUG:
			e100->force_rx_bug = 1;
			continue;

		case E100_EXTRA_RNR:
			e100->extra_rnr = 1;
			continue;

		case E100_HK_TIMEOUT:
			if (value != NULL) {
				e100->hk_timeout = strtoul(value, 0, 0);
			}
			continue;

		case TYPED_MEM:
			if (value != NULL)
				e100->tmem = value;
			continue;
		}

error:
		log(LOG_WARNING, "fxp%d: unknown option %s", instance, c);
		rc = EINVAL;
	}

	free(freeptr, M_TEMP);

	return rc;
}

static int e100_inter_entry_setup(e100_dev_t *e100)
{
	int rc = 0;
	if ((rc = interrupt_entry_init(&e100->inter, 0, NULL, e100->cfg.priority)) != EOK) {
		return rc;
	}
	e100->inter.func   = e100_process_interrupt;
	if (e100->kermask == 0) {
		e100->inter.enable = e100_enable_interrupt;
		e100->isrp = e100_isr;
	}
	else {
		e100->inter.enable = e100_enable_interrupt_kermask;
		e100->isrp = e100_isr_kermask;
	}
	e100->inter.arg = e100;

	return rc;
}

static void ifp_setup(e100_dev_t *e100, const char *ifname)
{
	struct ifnet *ifp;

	ifp = &e100->ecom.ec_if;
	ifp->if_softc = e100;
	strcpy(ifp->if_xname, ifname);
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX;
	if ( (e100->cfg.flags & NIC_FLAG_MULTICAST) ) {
		ifp->if_flags |= IFF_MULTICAST;
	}
	ifp->if_ioctl = e100_ioctl;
	ifp->if_start = e100_start;
	ifp->if_init = e100_init;
	ifp->if_stop = e100_stop;
	IFQ_SET_READY(&ifp->if_snd);

	strcpy((char *)e100->cfg.device_description, "Intel 8255x 10/100 Mbps");
	if_attach(ifp);
	ether_ifattach(ifp, e100->cfg.current_address);
}

static int e100_alloc(e100_dev_t *e100)
{
	e100->mem = (struct mem *)mmap(0, sizeof(struct mem), e100->prot_flags, e100->map_flags, e100->tmem_fd, 0);
	if (e100->mem == (struct mem *)MAP_FAILED) {
		return errno;
	}
	memset(e100->mem, 0, sizeof(struct mem));
	e100->dma_addr = (dma_addr_t)drvr_mphys(e100->mem);

	return 0;
}

static void e100_free(e100_dev_t *e100)
{
	if (e100->mem) {
		munmap(e100->mem, sizeof(struct mem));
		e100->mem = NULL;
	}
}

static int e100_attach(struct device *parent, struct device *self, void *aux)
{
	struct e100_dev			*idev;
	e100_dev_t				*e100;
	nic_config_t			*cfg;
	void					*head;
	struct pci_dev_info		*pci_info;
	int						rc;
	struct e100_attach_args	*e100_args;

	
	idev = (struct e100_dev *) self;
	head = idev->filler;
	e100 = NET_CACHELINE_ALIGN (head);
	idev->sc_e100 = e100;
	e100_args = aux;
	strcpy(e100->ecom.ec_if.if_xname, self->dv_xname);
	self->dv_dll_hdl = e100_args->dll_hdl;
	E100_DEBUG2(log(LOG_DEBUG, "%s: attach start", __devname__));

	e100->iopkt = iopkt_selfp;

	e100->iid = -1;

	cfg = &e100->cfg;

	/* set some defaults for the command line options */
	cfg->flags = NIC_FLAG_MULTICAST;
	cfg->device_index = e100_args->device_idx;
	pci_info = &e100->pci_info;
	memset ((char *)pci_info, 0, sizeof (struct pci_dev_info));
	pci_info->VendorId = cfg->vendor_id = e100_args->vendor_id;
	pci_info->DeviceId = cfg->device_id = e100_args->device_id;
	cfg->media_rate = cfg->duplex = -1;
	cfg->priority = IRUPT_PRIO_DEFAULT;
	cfg->iftype = IFT_ETHER;
	cfg->lan = -1;
	strcpy((char *)cfg->uptype, "en");

	e100->mdio_ctrl = mdio_ctrl_hw;
// #ifdef __X86__
// 	e100->use_io = 1;
// #else
	e100->use_io = 0;
// #endif
	if ((rc = e100_parse_options(e100, e100_args->options, e100_args->instance, cfg)) != 0) {
		log(LOG_ERR, "%s: error parsing options: %d", __devname__, rc);
		e100_destroy(e100, ATTACH_DONE);
		return rc;
	}
	E100_DEBUG(log(LOG_DEBUG, "%s: CBS {%d %d %d}", __devname__, e100->params.cbs.min, e100->params.cbs.max, e100->params.cbs.count));
	E100_DEBUG(log(LOG_DEBUG, "%s: RFDS {%d %d %d}", __devname__, e100->params.rfds.min, e100->params.rfds.max, e100->params.rfds.count));

	cfg->lan = idev->sc_dev.dv_unit;

	if (e100->cfg.mtu == 0 || e100->cfg.mtu > ETH_MAX_PKT_LEN)
		e100->cfg.mtu = ETH_MAX_PKT_LEN;

	if (e100->cfg.mru == 0 || e100->cfg.mru > ETH_MAX_PKT_LEN)
		e100->cfg.mru = ETH_MAX_PKT_LEN;

	e100->cfg.revision = NIC_CONFIG_REVISION;

	if ( (rc = nic_force_advertise(e100)) != 0 ) {
		e100_destroy(e100, ATTACH_DONE);
		return rc;
	}

	if (nw_pci_hdl == -1) {
		log(LOG_ERR, "%s: unable to locate pci server", __devname__);
		e100_destroy(e100, ATTACH_DONE);
		return EAGAIN;
	}
	
	setup_tmem(e100);
	setup_mmap_flags(e100);

	if ( (rc = pci_enable_device(e100)) ) {
		log(LOG_ERR, "%s: pci_enable_device: %d", __devname__, rc);
		e100_destroy(e100, ATTACH_DONE);
		return rc;
	}

	if (cfg->num_irqs == 0) {
		cfg->irq[0] = pci_info->Irq;
		cfg->num_irqs = 1;
	}

	if (cfg->num_io_windows == 0) {
		cfg->io_window_base[0] = PCI_IO_ADDR(pci_info->CpuBaseAddress[1]);
		cfg->io_window_size[0] = pci_info->BaseAddressSize[1];
		cfg->num_io_windows = 1;
	}

	if (cfg->num_mem_windows == 0) {
		cfg->mem_window_base[0] = PCI_MEM_ADDR(pci_info->CpuBaseAddress[0]);
		cfg->mem_window_size[0] = pci_info->BaseAddressSize[0];
		cfg->num_mem_windows = 1;
	}

	if (pci_info->PciRom)
		cfg->rom_base = PCI_ROM_ADDR(pci_info->CpuRom);

	e100->bmstr = pci_info->CpuBmstrTranslation;

	evcnt_attach_dynamic(&e100->ev_txdrop, EVCNT_TYPE_MISC, NULL, self->dv_xname, "txdrop");

	e100->cachectl.fd = NOFD;
	if (cache_init(0, &e100->cachectl, NULL) == -1) {
		rc = errno;
		log(LOG_ERR, "%s: cache_init: %d", __devname__, rc);
		e100_destroy(e100, EVCNT_DONE);
		return rc;
	}

	if (!e100->use_io) {
		e100->csr = (struct csr *)mmap_device_memory(NULL,
					/*sizeof(struct csr)*/ e100->cfg.mem_window_size[0],
					e100->prot_flags | PROT_NOCACHE, 0,
					e100->cfg.mem_window_base[0]);
		
		if (e100->csr == (struct csr *)MAP_FAILED) {
			rc = errno;
			log(LOG_ERR,
			    "%s: Unable to mmap_device_memory: %d.", __devname__, rc);
			e100_destroy(e100, CACHE_INIT_DONE);
			return rc;
		}
	} else {
		e100->csr = (struct csr *)mmap_device_io(sizeof(struct csr),
		    e100->cfg.io_window_base[0]);

		if (e100->csr == (struct csr *)MAP_DEVICE_FAILED) {
			rc = errno;
			log(LOG_ERR,
			    "%s: Unable to mmap_device_io: %d.", __devname__, rc);
			e100_destroy(e100, CACHE_INIT_DONE);
			return rc;
		}
	}

	if (e100_args->driver_data)
		e100->flags |= ich;
	else
		e100->flags &= ~ich;

	/* Ethernet stats we are interested in */
	e100->stats.un.estats.valid_stats =
	    NIC_ETHER_STAT_XCOLL_ABORTED |
	    NIC_ETHER_STAT_SINGLE_COLLISIONS |
	    NIC_ETHER_STAT_MULTI_COLLISIONS |
		NIC_ETHER_STAT_LATE_COLLISIONS |
		NIC_ETHER_STAT_NO_CARRIER |
		NIC_ETHER_STAT_TX_DEFERRED |
		NIC_ETHER_STAT_INTERNAL_TX_ERRORS |
		NIC_ETHER_STAT_TOTAL_COLLISION_FRAMES |
		NIC_ETHER_STAT_SHORT_PACKETS |
		NIC_ETHER_STAT_OVERSIZED_PACKETS |
		NIC_ETHER_STAT_FCS_ERRORS |
		NIC_ETHER_STAT_ALIGN_ERRORS |
		NIC_ETHER_STAT_INTERNAL_RX_ERRORS;
	/* Generic networking stats we are interested in */
	e100->stats.valid_stats =
	    NIC_STAT_TX_FAILED_ALLOCS | NIC_STAT_RX_FAILED_ALLOCS |
	     NIC_STAT_TXED_BROADCAST |
	     NIC_STAT_RXED_BROADCAST;
	if ( (cfg->flags & NIC_FLAG_MULTICAST) ) {
		e100->stats.valid_stats |= NIC_STAT_TXED_MULTICAST | NIC_STAT_RXED_MULTICAST;
	}


// 	memset(&e100->cmd_lock, 0, sizeof(e100->cmd_lock));
// 	memset(&e100->cb_lock, 0, sizeof(e100->cb_lock));
	atomic_clr(&e100->irq_sw_gen, 0xFFFFFFFF);
	atomic_clr(&e100->stopping, 0xFFFFFFFF);

	/* Reset the device before pci_set_master in case device is in some
	 * funky state and has an interrupt pending - hint: we don't have the
	 * interrupt handler registered yet. */
	e100_hw_reset(e100);

// 	if ( (rc = pci_set_master(e100)) ) {
// 		log(LOG_ERR, "%s: pci_set_master: %d", __devname__, rc);
// 		e100_destroy(e100, MMAP_CSR_DONE);
// 		return rc;
// 	}

	if ( (rc = e100_eeprom_load(e100)) ) {
		e100_destroy(e100, PCI_MASTER_DONE);
		return rc;
	}

	e100_get_defaults(e100);

	/* D100 MAC doesn't allow rx of vlan packets with normal MTU */
	if (e100->mac >= mac_82558_D101_A4)
		e100->ecom.ec_capabilities |= ETHERCAP_VLAN_MTU;

	if ((rc = e100_alloc(e100))) {
		log(LOG_ERR, "%s: Cannot alloc driver memory, aborting", __devname__);
		e100_destroy(e100, PCI_MASTER_DONE);
		return rc;
	}
	
	if ( (rc = e100_findphy(e100)) ) {
		e100_destroy(e100, MMAP_E100_DONE);
		return rc;
	}
	/* Copy MAC address from EEPROM */
	memcpy(e100->cfg.permanent_address, e100->eeprom, ETH_MAC_LEN);
	E100_DEBUG(log(LOG_DEBUG, "%s: eeprom MAC:%02X:%02X:%02X:%02X:%02X:%02X", __devname__,
					e100->cfg.permanent_address[0], e100->cfg.permanent_address[1],
					e100->cfg.permanent_address[2], e100->cfg.permanent_address[3],
					e100->cfg.permanent_address[4], e100->cfg.permanent_address[5]));
	if ( !valid_mac_addr((char *)e100->cfg.permanent_address) ) {
		if (!e100->eeprom_bad_csum_allow) {
			log(LOG_ERR, "%s: Invalid MAC address from EEPROM, aborting", __devname__);
			e100_destroy(e100, MMAP_E100_DONE);
		} else {
			log(LOG_WARNING, "%s: Invalid MAC address from EEPROM, using random.", __devname__);
			fill_random_mac((char *)e100->cfg.permanent_address);
		}
	}
	/* Override with MAC address from syspage, if there is one */
	nic_get_syspage_mac((char *)e100->cfg.permanent_address);

	/* check for command line override */
	if ( !valid_mac_addr((char *)e100->cfg.current_address) )
		memcpy(e100->cfg.current_address, e100->cfg.permanent_address, ETH_MAC_LEN);

// FIXME needed there or in _init()?
	if ( (rc = e100_alloc_cbs(e100)) ) {
		log(LOG_ERR,
			"%s: Unable to alloc cbs: %d.", __devname__, rc);
		e100_destroy(e100, MMAP_E100_DONE);
		return rc;
	}

	if ( (rc = e100_rx_alloc_list(e100)) ) {
		log(LOG_ERR,
			"%s: Unable to alloc rxs: %d.", __devname__, rc);
		e100_destroy(e100, ALLOC_CBS_DONE);
		return rc;
	}

	if ((rc = e100_inter_entry_setup(e100)) != 0) {
		e100_destroy(e100, ALLOC_RXS_DONE);
		return rc;
	}

	// hook up so media devctls work
	bsd_mii_initmedia(e100);

	ifp_setup(e100, self->dv_xname);

	if ((e100->sd_hook = shutdownhook_establish(e100_shutdown,
	    e100)) == NULL) {
		e100_destroy(e100, IF_SETUP_DONE);
		return ENOMEM;
	}

	if ( e100->cfg.verbose ) {
		log(LOG_INFO, "%s:", __devname__);
		nic_dump_config(&e100->cfg);
	}

	return 0;
}

static void e100_shutdown(void *arg)
{
	/* All of io-pkt is going away.  Just quiet hardware. */

	e100_destroy((e100_dev_t *)arg, DESTROY_ALL);
}

static int e100_detach(struct device *dev, int flags)
{
	struct e100_dev	*kdev;
	e100_dev_t		*e100;

	kdev = (struct e100_dev *)dev;
	e100 = kdev->sc_e100;

	e100_destroy(e100, DESTROY_ALL);

	return 0;
}


static pthread_mutex_t e100_destroy_mutex = PTHREAD_MUTEX_INITIALIZER;
static void e100_destroy(e100_dev_t *e100, enum destroy_variants dstage)
{
	struct	ifnet		*ifp;

	ifp = &e100->ecom.ec_if;

	E100_DEBUG(log(LOG_DEBUG, "%s: destroy with stage %d", __devname__, (int)dstage));

	pthread_mutex_lock(&e100_destroy_mutex);
	if (e100->dying) {
		goto destroy_done;
	}
	switch (dstage) {
		case DESTROY_ALL:
		/* Don't init() while we're dying. */
			e100->dying = 1;
			e100_stop(ifp, 1);

		case IF_SETUP_DONE:
			ether_ifdetach(ifp);
			if_detach(ifp);

		case INTER_ENTRY_DONE:
			if (ISSTACK) {
				if ( e100->inter.func ) {
					interrupt_entry_remove(&e100->inter, NULL);	/* Must be 'the stack' to call this */
				}
			} else {
				log(LOG_ERR, "%s: stop: unexpected context", __devname__);
			}
		case ALLOC_RXS_DONE:
			e100_rx_clean_list(e100, 1);

		case ALLOC_CBS_DONE:
			e100_clean_cbs(e100, 1);

		case MMAP_E100_DONE:
			e100_free(e100);

		case PCI_MASTER_DONE:
// 			pci_disable_device(e100);

		case MMAP_CSR_DONE:
			if (e100->use_io) {
				munmap_device_io((uintptr_t)e100->csr, sizeof(struct csr));
			} else {
				munmap_device_memory((void *)e100->csr, sizeof(struct csr));
			}

		case CACHE_INIT_DONE:
			cache_fini(&e100->cachectl);

		case EVCNT_DONE:
			evcnt_detach (&e100->ev_txdrop);
		
		case PCI_ENABLE_DONE:
			pci_disable_device(e100);

		case ATTACH_DONE:
			break;
	}

	if (e100->sd_hook != NULL)
		shutdownhook_disestablish (e100->sd_hook);

destroy_done:
	pthread_mutex_unlock(&e100_destroy_mutex);
}

static void e100_hk_callout(void *arg)
{
	e100_dev_t				*e100   = arg;
	struct ifnet			*ifp      = &e100->ecom.ec_if;
	struct nw_work_thread	*wtp      = WTP;

	/* Software generated interrupt to recover from (rare) Rx
	 * allocation failure.
	 * Unfortunately have to use a spinlock to not re-enable interrupts
	 * accidentally, due to hardware that shares a register between the
	 * interrupt mask bit and the SW Interrupt generation bit */
// 	InterruptLock(&e100->cmd_lock);
	atomic_set(&e100->irq_sw_gen, irq_sw_gen);
// 	iowrite8(ioread8(&e100->csr->scb.cmd_hi) | irq_sw_gen, &e100->csr->scb.cmd_hi);
// 	e100_write_flush(e100);
// 	InterruptUnlock(&e100->cmd_lock);

	e100_update_stats(e100);

	NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt_selfp, wtp);
	if (e100->mac <= mac_82557_D100_C || e100->force_rx_bug)
		/* Issue a multicast command to workaround a 557 lock up */
		e100_set_multicast_list(e100);
	e100_tx_clean(e100);
	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() stat_ack %02X status %02X mask %02X", __devname__, __func__, ioread8(&e100->csr->scb.stat_ack), ioread8(&e100->csr->scb.status), ioread8(&e100->csr->scb.cmd_hi));
	)
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt_selfp, wtp);

	callout_msec(&e100->hk_callout, e100->hk_timeout * 1000, e100_hk_callout, e100);
}

int e100_init(struct ifnet *ifp)
{
	int ret;
	e100_dev_t *e100;
	struct _iopkt_self *iopkt;
	struct nw_work_thread *wtp;

	/*
	 * - enable hardware.
	 *   - look at ifp->if_capenable_[rx/tx]
	 *   - enable promiscuous / multicast filter.
	 * - attach to interrupt.
	 */
	e100 = (e100_dev_t *)ifp->if_softc;
	if (e100->dying == 1)
		return 0;
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() start", __devname__, __func__));
	iopkt = e100->iopkt;
	wtp = WTP;

	e100_stop(ifp, 0);

	NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	ifp->if_flags_tx |= IFF_OACTIVE;

	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs %p", __devname__, __func__, e100->cbs));
	if ((ret = e100_rx_init_list(e100, wtp))) {
		log(LOG_ERR, "%s: Receive list init is failed: %d", __devname__, ret);
		goto do_err;
	}
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs0 %p", __devname__, __func__, e100->cbs));
	if ((ret = e100_hw_init(e100))) {
		log(LOG_ERR, "%s: Hardware init is failed: %d", __devname__, ret);
		goto err_rx_clean_list;
	}
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs1 %p", __devname__, __func__, e100->cbs));
	e100_set_multicast_list(e100);
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs2 %p", __devname__, __func__, e100->cbs));
	e100_start_receiver(e100, NULL);
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs3 %p", __devname__, __func__, e100->cbs));
	callout_msec (&e100->mii_callout, 3 * 1000, e100_mii_callout, e100);
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs4 %p", __devname__, __func__, e100->cbs));
	// start housekeeping callout - give it 10 seconds
	callout_msec(&e100->hk_callout, e100->hk_timeout * 1000, e100_hk_callout, e100);
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs5 %p", __devname__, __func__, e100->cbs));
	if (e100->iid == -1) {
		if ((ret = InterruptAttach_r(e100->cfg.irq[0], e100->isrp,
		    e100, sizeof(*e100), _NTO_INTR_FLAGS_TRK_MSK)) < 0) {
			ret = -ret;
			log(LOG_ERR, "%s: Interrupt attach for %X failed", __devname__, e100->cfg.irq[0]);
			goto do_err;
		}
		e100->iid = ret;
	}
	atomic_clr(&e100->stopping, 0xFFFFFFFF);
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs6 %p", __devname__, __func__, e100->cbs));
	e100_enable_irq(e100);
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs7 %p e100 %p", __devname__, __func__, e100->cbs, e100));
// 	get mtu from stack, mostly for nicinfo
	e100->cfg.mtu = ifp->if_mtu;
	e100->cfg.mru = ifp->if_mtu;
	ifp->if_flags_tx |= IFF_RUNNING;
	ifp->if_flags_tx &= ~IFF_OACTIVE;
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	ifp->if_flags |= IFF_RUNNING;
	return EOK;

err_rx_clean_list:
	e100_rx_clean_list(e100, 0);

do_err:
	ifp->if_flags_tx &= ~IFF_OACTIVE;
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);

	log(LOG_ERR, "%s: not running.", __devname__);
	return ret;
}

void e100_stop(struct ifnet *ifp, int disable)
{
	e100_dev_t			*e100 = ifp->if_softc;
	struct _iopkt_self		*iopkt = e100->iopkt;
	struct nw_work_thread	*wtp = WTP;
	int i;
	/*
	 * - Cancel any pending io
	 * - Clear any interrupt source registers
	 * - Clear any interrupt pending registers
	 * - Release any queued transmit buffers.
	 */
	atomic_set(&e100->stopping, 0x1);
	i = 0;

	e100_disable_irq(e100);

	E100_DEBUG(log(LOG_DEBUG, "%s: %s() start", __devname__, __func__));
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs %p", __devname__, __func__, e100->cbs));
	/* Shut off the housekeeping & mii callout */
	callout_stop(&e100->hk_callout);
	callout_stop(&e100->mii_callout);

	if (e100->mdi) {
		MDI_DeRegister((mdi_t **)&e100->mdi);
		e100->mdi = NULL;
	}

	e100_hw_reset(e100);

	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs2 %p", __devname__, __func__, e100->cbs));
	/* Lock out the transmit side */
	NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	for (i = 0; i < 10; i++) {
		if ((ifp->if_flags_tx & IFF_OACTIVE) == 0)
			break;
		NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
		delay(50);
		NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	}

	if (i < 10) {
		ifp->if_flags_tx &= ~IFF_RUNNING;
		NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	}
	else {
		/* Heavy load or bad luck.  Try the big gun. */
		quiesce_all();
		ifp->if_flags_tx &= ~IFF_RUNNING;
		unquiesce_all();
	}
	e100_clean_cbs(e100, 0);

	if (disable) {
		if (e100->iid != -1) {
			InterruptDetach(e100->iid);
			e100->iid = -1;
		}
		e100_rx_clean_list(e100, 0);
	}
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs3 %p", __devname__, __func__, e100->cbs));
	E100_DEBUG(log(LOG_DEBUG, "%s: %s() e100->cbs4 %p", __devname__, __func__, e100->cbs));
	/* Mark the interface as down */
	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);
}
