/*
 * (c) 2016-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller MII PHY routines                                       *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"

void	e100_mdi_callback (void *hdl, uint8_t phy_id, uint8_t link_state)
{
	e100_dev_t	*e100 = (e100_dev_t *)hdl;
	int				i, mode;
	char			*s;
	struct ifnet	*ifp = &e100->ecom.ec_if;

	switch (link_state) {
		case MDI_LINK_UP:
			if ((i = MDI_GetActiveMedia(e100->mdi, e100->cfg.phy_addr, &mode)) != MDI_LINK_UP) {
				if (e100->cfg.verbose > 8)
					slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s: callback GetActiveMedia returned %x", __devname__, i);
				mode = 0;
			}

			switch (mode) {
				case MDI_10bTFD:
					s = "10 BaseT Full Duplex";
					e100->cfg.duplex = 1;
					e100->cfg.media_rate = 10000L;
					break;
				case MDI_10bT:
					s = "10 BaseT Half Duplex";
					e100->cfg.duplex = 0;
					e100->cfg.media_rate = 10000L;
					break;
				case MDI_100bTFD:
					s = "100 BaseT Full Duplex";
					e100->cfg.duplex = 1;
					e100->cfg.media_rate = 100000L;
					break;
				case MDI_100bT:
					s = "100 BaseT Half Duplex";
					e100->cfg.duplex = 0;
					e100->cfg.media_rate = 100000L;
					break;
				case MDI_100bT4:
					s = "100 BaseT4";
					e100->cfg.duplex = 0;
					e100->cfg.media_rate = 100000L;
					break;
				case MDI_1000bT:
					s = "1000 BaseT Half Duplex";
					e100->cfg.duplex = 0;
					e100->cfg.media_rate = 1000000L;
					break;
				case MDI_1000bTFD:
					s = "1000 BaseT Full Duplex";
					e100->cfg.duplex = 1;
					e100->cfg.media_rate = 1000000L;
					break;
				default:
					s = "Unknown";
					e100->cfg.duplex = 0;
					e100->cfg.media_rate = 0L;
					break;
			}

			if (e100->cfg.verbose) {
				log(LOG_INFO, "%s: Link up (%s)", __devname__, s);
			}

			e100->cfg.flags &= ~NIC_FLAG_LINK_DOWN;
			if_link_state_change(ifp, LINK_STATE_UP);

			if (mode)
				e100_adjust_adaptive_ifs(e100, e100->cfg.media_rate / 1000, e100->cfg.duplex);
			break;
		case MDI_LINK_DOWN:
			e100->cfg.media_rate = e100->cfg.duplex = -1;
	//		MDI_ResetPhy (rtl->mdi, rtl->cfg.phy_addr, WaitBusy);
	//		MDI_SyncPhy (rtl->mdi, rtl->cfg.phy_addr);
			MDI_AutoNegotiate (e100->mdi, e100->cfg.phy_addr, NoWait);
			e100->cfg.flags |= NIC_FLAG_LINK_DOWN;

			if (e100->cfg.verbose) {
				slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s: Link down %d", __devname__, e100->cfg.lan);
			}

			if_link_state_change(ifp, LINK_STATE_DOWN);
			break;
		default:
			if (e100->cfg.verbose) {
				log(LOG_INFO, "%s: Unknown link state %hhu", __devname__, link_state);
			}
			break;
	}
}

void e100_mii_callout(void *arg)
{
	e100_dev_t	*e100   = arg;
	if ((e100->cfg.flags & NIC_FLAG_LINK_DOWN) ||
	   (e100->cfg.media_rate <= 0) ||
	   !e100->pkts_received) {
		if (e100->cfg.verbose > 12)
			log(LOG_DEBUG, "%s: calling MDI_MonitorPhy()", __devname__);
		MDI_MonitorPhy (e100->mdi);
	}
	e100->pkts_received = 0;
	callout_msec(&e100->mii_callout, 3 * 1000, e100_mii_callout, e100);
}

uint16_t mii_read(void *hdl, uint8_t phy_id, uint8_t reg_num)
{
	e100_dev_t		*e100 = (e100_dev_t *)hdl;
	return e100->mdio_ctrl(e100, phy_id, mdi_read, reg_num, 0);
}

void mii_write(void *hdl, uint8_t phy_id, uint8_t reg_num, uint16_t val)
{
	e100_dev_t		*e100 = (e100_dev_t *)hdl;
	e100->mdio_ctrl(e100, phy_id, mdi_write, reg_num, val);
}
/* the standard mdio_ctrl() function for usual MII-compliant hardware */
u16 mdio_ctrl_hw(e100_dev_t *e100, u32 addr, u32 dir, u32 reg, u16 data)
{
	u32 data_out = 0;
	unsigned int i;

	for (i = 100; i; --i) {
		if (ioread32(&e100->csr->mdi_ctrl) & mdi_ready)
			break;
		udelay(20);
	}
	if (unlikely(!i)) {
		log(LOG_DEBUG, "%s: e100.mdio_ctrl won't go Ready", __devname__);
		return 0;		/* No way to indicate timeout error */
	}
	iowrite32((reg << 16) | (addr << 21) | dir | data, &e100->csr->mdi_ctrl);

	for (i = 0; i < 100; i++) {
		udelay(20);
		if ((data_out = ioread32(&e100->csr->mdi_ctrl)) & mdi_ready)
			break;
	}
	E100_DEBUG2(
		log(LOG_DEBUG, "%s: %s:addr=%d, reg=%d, data_in=0x%04X, data_out=0x%04X", __devname__,
		     dir == mdi_read ? "READ" : "WRITE",
		     addr, reg, data, data_out));
	E100_DEBUG2(log(LOG_DEBUG, "%s: mdi_ctrl %08X BIT29 : %d", __devname__, ioread32(&e100->csr->mdi_ctrl), (ioread32(&e100->csr->mdi_ctrl) & (1<<29)));)
	return (u16)data_out;
}

/* slightly tweaked mdio_ctrl() function for phy_82552_v specifics */
static u16 mdio_ctrl_phy_82552_v(e100_dev_t *e100,
				 u32 addr,
				 u32 dir,
				 u32 reg,
				 u16 data)
{
	if ((reg == MDI_BMCR) && (dir == mdi_write)) {
		if (data & (BMCR_RESTART_AN | BMCR_AN_ENABLE)) {
			u16 advert = mii_read(e100, e100->cfg.phy_addr,
							MDI_ANAR);

			/*
			 * Workaround Si issue where sometimes the part will not
			 * autoneg to 100Mbps even when advertised.
			 */
			if (advert & ANAR_100bTFD)
				data |= BMCR_SPEED_100 | BMCR_FULL_DUPLEX;
			else if (advert & ANAR_100bT)
				data |= BMCR_SPEED_100;
		}
	}
	return mdio_ctrl_hw(e100, addr, dir, reg, data);
}

/* Fully software-emulated mdio_ctrl() function for cards without
 * MII-compliant PHYs.
 * For now, this is mainly geared towards 80c24 support; in case of further
 * requirements for other types (i82503, ...?) either extend this mechanism
 * or split it, whichever is cleaner.
 */
static u16 mdio_ctrl_phy_mii_emulated(e100_dev_t *e100,
				      u32 addr,
				      u32 dir,
				      u32 reg,
				      u16 data)
{
	/* might need to allocate a e100_priv'ed register array eventually
	 * to be able to record state changes, but for now
	 * some fully hardcoded register handling ought to be ok I guess. */

	if (dir == mdi_read) {
		switch (reg) {
		case MDI_BMCR:
			/* Auto-negotiation, right? */
			return  BMCR_AN_ENABLE |
				BMCR_FULL_DUPLEX;
		case MDI_BMSR:
			return	BMSR_LINK_STATUS /* for mii_link_ok() */ |
				BMSR_AN_ABILITY |
				BMSR_10FD;
		case MDI_ANAR:
			/* 80c24 is a "combo card" PHY, right? */
			return	ANAR_10bT |
				ANAR_10bTFD;
		default:
			log(LOG_DEBUG, "%s: %s:addr=%d, reg=%d, data=0x%04X: unimplemented emulation!", __devname__,
				     dir == mdi_read ? "READ" : "WRITE",
				     addr, reg, data);
			return 0xFFFF;
		}
	} else {
		switch (reg) {
		default:
			log(LOG_DEBUG, "%s: %s:addr=%d, reg=%d, data=0x%04X: unimplemented emulation!", __devname__,
				     dir == mdi_read ? "READ" : "WRITE",
				     addr, reg, data);
			return 0xFFFF;
		}
	}
}
int e100_phy_supports_mii(e100_dev_t *e100)
{
	/* for now, just check it by comparing whether we
	   are using MII software emulation.
	*/
	return (e100->mdio_ctrl != mdio_ctrl_phy_mii_emulated);
}

int e100_phy_check_without_mii(e100_dev_t *e100)
{
	u8 phy_type;
	int without_mii;

	phy_type = (e100->eeprom[eeprom_phy_iface] >> 8) & 0x0f;

	switch (phy_type) {
	case NoSuchPhy: /* Non-MII PHY; UNTESTED! */
	case I82503: /* Non-MII PHY; UNTESTED! */
	case S80C24: /* Non-MII PHY; tested and working */
		/* paragraph from the FreeBSD driver, "FXP_PHY_80C24":
		 * The Seeq 80c24 AutoDUPLEX(tm) Ethernet Interface Adapter
		 * doesn't have a programming interface of any sort.  The
		 * media is sensed automatically based on how the link partner
		 * is configured.  This is, in essence, manual configuration.
		 */
		log(LOG_DEBUG, "%s: found MII-less i82503 or 80c24 or other PHY", __devname__);

		e100->mdio_ctrl = mdio_ctrl_phy_mii_emulated;
		e100->cfg.phy_addr = 0; /* is this ok for an MII-less PHY? */

		/* these might be needed for certain MII-less cards...
		 * e100->flags |= ich;
		 * e100->flags |= ich_10h_workaround; */

		without_mii = 1;
		break;
	default:
		without_mii = 0;
		break;
	}
	return without_mii;
}
#define MII_RESV1		0x17	/* Reserved...                 */
#define MII_TPISTATUS	0x1b	/* TPI status for 10mbps       */
#define MII_NCONFIG		0x1c	/* Network interface config    */
#define NCONFIG_AUTO_SWITCH	0x0080
#define MII_NSC_CONG		MII_RESV1
#define NSC_CONG_ENABLE		0x0100
#define NSC_CONG_TXREADY	0x0400
#define ADVERTISE_FC_SUPPORTED	0x0400
int e100_findphy (e100_dev_t *e100)

{
	int				an_capable,status;
	int				addr;
	uint16_t		reg;
	uint16_t		bmcr, stat, cong, id_lo, id_hi;

	e100->cfg.phy_addr = -1;

	if (e100->mdi) {
		MDI_DeRegister ((mdi_t **)&e100->mdi);
	}

	status = MDI_Register_Extended (e100, mii_write, mii_read,
		e100_mdi_callback, (mdi_t **)&e100->mdi, NULL, 0, 0);
	
	if (status != MDI_SUCCESS) {
		log(LOG_ERR, "%s: Cannot register the mii routines", __devname__);
		e100->mdi = NULL;
		return -1;
	}
	
	callout_init(&e100->mii_callout);

	for (e100->cfg.phy_addr = 0;
	    e100->cfg.phy_addr < 32; e100->cfg.phy_addr++) {
#if 0
		if (MDI_FindPhy(e100->mdi, e100->cfg.phy_addr) == MDI_SUCCESS)
			break;
#else
		bmcr = mii_read(e100, e100->cfg.phy_addr, MDI_BMCR);
		stat = mii_read(e100, e100->cfg.phy_addr, MDI_BMSR);
		stat = mii_read(e100, e100->cfg.phy_addr, MDI_BMSR);
		if (!((bmcr == 0xFFFF) || ((stat == 0) && (bmcr == 0))))
			break;
#endif
	}
	addr = e100->cfg.phy_addr;
///////////
	if (addr == 32) {
		/* uhoh, no PHY detected: check whether we seem to be some
		 * weird, rare variant which is *known* to not have any MII.
		 * But do this AFTER MII checking only, since this does
		 * lookup of EEPROM values which may easily be unreliable. */
		if (e100_phy_check_without_mii(e100))
			return 0; /* simply return and hope for the best */
		else {
			/* for unknown cases log a fatal error */
			log(LOG_ERR, "%s: Cannot find an active PHY", __devname__);
			return -1;
		}
	} else if (e100->cfg.verbose) {
		log(LOG_INFO, "%s: MII transceiver found at address %d.", __devname__, e100->cfg.phy_addr);
	}
	
	if (MDI_InitPhy(e100->mdi, e100->cfg.phy_addr) != MDI_SUCCESS) {
		log(LOG_ERR, "%s: Cannot init the PHY status", __devname__);
		return -1;
	}

	/* Get phy ID */
	id_lo = mii_read(e100, e100->cfg.phy_addr, MDI_PHYID_1);
	id_hi = mii_read(e100, e100->cfg.phy_addr, MDI_PHYID_2);
	e100->phy = (u32)id_hi << 16 | (u32)id_lo;
	if (e100->cfg.verbose) {
		log(LOG_INFO, "%s: phy ID = 0x%08X", __devname__,  e100->phy);
	}

	/* Select the phy and isolate the rest */
	for (addr = 0; addr < 32; addr++) {
		if (addr != e100->cfg.phy_addr) {
			mii_write(e100, addr, MDI_BMCR, BMCR_ISOLATE);
		} else if (e100->phy != phy_82552_v) {
			bmcr = mii_read(e100, addr, MDI_BMCR);
			mii_write(e100, addr, MDI_BMCR,
				bmcr & ~BMCR_ISOLATE);
		}
	}
	/*
	 * Workaround for 82552:
	 * Clear the ISOLATE bit on selected phy_id last (mirrored on all
	 * other phy_id's) using bmcr value from addr discovery loop above.
	 */
	if (e100->phy == phy_82552_v)
		mii_write(e100, e100->cfg.phy_addr, MDI_BMCR,
			bmcr & ~BMCR_ISOLATE);

	/* Handle National tx phys */
#define NCS_PHY_MODEL_MASK	0xFFF0FFFF
	if ((e100->phy & NCS_PHY_MODEL_MASK) == phy_nsc_tx) {
		/* Disable congestion control */
		cong = mii_read(e100, e100->cfg.phy_addr, MII_NSC_CONG);
		cong |= NSC_CONG_TXREADY;
		cong &= ~NSC_CONG_ENABLE;
		mii_write(e100, e100->cfg.phy_addr, MII_NSC_CONG, cong);
	}

	if (e100->phy == phy_82552_v) {
		u16 advert = mii_read(e100, e100->cfg.phy_addr, MDI_ANAR);

		/* assign special tweaked mdio_ctrl() function */
		e100->mdio_ctrl = mdio_ctrl_phy_82552_v;

		/* Workaround Si not advertising flow-control during autoneg */
		advert |= ANAR_FLOW_CONTROL | ANAR_FLOW_ASYMMETRIC;
		mii_write(e100, e100->cfg.phy_addr, MDI_ANAR, advert);

		/* Reset for the above changes to take effect */
		bmcr = mii_read(e100, e100->cfg.phy_addr, MDI_BMCR);
		bmcr |= BMCR_RESET;
		mii_write(e100, e100->cfg.phy_addr, MDI_BMCR, bmcr);
	} else if ((e100->mac >= mac_82550_D102) || ((e100->flags & ich) &&
	   (mii_read(e100, e100->cfg.phy_addr, MII_TPISTATUS) & 0x8000) &&
		(e100->eeprom[eeprom_cnfg_mdix] & eeprom_mdix_enabled))) {
		/* enable/disable MDI/MDI-X auto-switching. */
		mii_write(e100, e100->cfg.phy_addr, MII_NCONFIG,
				e100->force_advertise ? 0 : NCONFIG_AUTO_SWITCH);
	}
//////////////////////////////////////////

	e100->cfg.connector = NIC_CONNECTOR_MII;
	an_capable = mii_read(e100, e100->cfg.phy_addr, MDI_BMSR) & 8;

    //
    // if the user has specified the speed or duplex
    // or if the phy cannot auto-negotiate ...
    //
	if (e100->force_advertise != -1 || !an_capable) {
		reg = mii_read(e100, e100->cfg.phy_addr, MDI_BMCR);

		reg &= ~(BMCR_RESTART_AN|BMCR_SPEED_100|BMCR_FULL_DUPLEX);

		if (an_capable && e100->force_advertise != 0) {
			/*
			 * If we force the speed, but the link partner
			 * is autonegotiating, there is a greater chance
			 * that everything will work if we advertise with
			 * the speed that we are forcing to.
			 */
			MDI_SetAdvert(e100->mdi,
			    e100->cfg.phy_addr, e100->force_advertise);

			reg |= BMCR_RESTART_AN | BMCR_AN_ENABLE;

			if (e100->cfg.verbose)
				log(LOG_INFO, "%s: "
				    "restricted autonegotiate (%dMbps only)", __devname__,
				    e100->cfg.media_rate/1000);
		} else {
			reg &= ~BMCR_AN_ENABLE;

			if (e100->cfg.verbose)
				log(LOG_INFO, "%s: forcing the link", __devname__);
		}

		if (e100->cfg.duplex > 0)
			reg |= BMCR_FULL_DUPLEX;
		if (e100->cfg.media_rate == 100*1000)
			reg |= BMCR_SPEED_100;

		mii_write(e100, e100->cfg.phy_addr, MDI_BMCR, reg);

		if (reg & BMCR_AN_ENABLE)
			MDI_EnableMonitor(e100->mdi, 1);

	} else {	// normal auto-negotiation mode

		e100->cfg.flags |= NIC_FLAG_LINK_DOWN;
		MDI_AutoNegotiate(e100->mdi, e100->cfg.phy_addr, NoWait);
		status = MDI_EnableMonitor(e100->mdi, 1);

		if (status != MDI_SUCCESS)
			log(LOG_ERR,
			    "%s: MDI_EnableMonitor returned %x", __devname__, status);
	}
	
	return (0);
}

int nic_force_advertise(e100_dev_t *e100)
{
	nic_config_t *cfg;

	cfg = &e100->cfg;
	e100->force_advertise = -1;

	// did user specify either of speed or duplex on the cmd line?
	if ((cfg->media_rate != -1) || (cfg->duplex != -1)) {

		if (cfg->media_rate == -1) {
			log(LOG_ERR, "%s(): must also specify speed when duplex is specified", __FUNCTION__);
			return EINVAL;
		}
		if (cfg->duplex == -1) {
			log(LOG_ERR, "%s(): must also specify duplex when speed is specified", __FUNCTION__);
			return EINVAL;
		}

		// we get here, we know both media_rate and duplex are set

		switch(cfg->media_rate) {
			case 0:
				e100->force_advertise = 0;  // disable link
				break;
			case 10*1000:
				e100->force_advertise = cfg->duplex ? MDI_10bTFD : MDI_10bT;
				break;
			case 100*1000:
				e100->force_advertise = cfg->duplex ? MDI_100bTFD : MDI_100bT;
				break;
			default:
				log(LOG_ERR, "%s(): invalid speed: %d", __FUNCTION__, cfg->media_rate/1000);
				return EINVAL;
		}
	}

	return 0;
}
