/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller receive routines                                       *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"

void e100_rx_clean_list(e100_dev_t *e100, int full_clean)
{
	struct rx *rx;
	unsigned int i, count = e100->params.rfds.count;

	e100->ru_running = RU_UNINITIALIZED;

	if (e100->rxs) {
		for (rx = e100->rxs, i = 0; i < count; rx++, i++) {
			if (rx->m) {
				m_freem(rx->m);
				rx->m = NULL;
			}
		}
		if ( full_clean ) {
			free(e100->rxs, M_DEVBUF);
			e100->rxs = NULL;
		} else {
			memset(e100->rxs, 0x00, sizeof(struct rx) * count);
		}
	}

	e100->rx_to_use = e100->rx_to_clean = NULL;
}

int e100_rx_alloc_list(e100_dev_t *e100)
{
	unsigned int count = e100->params.rfds.count;

	e100->rxs = malloc(sizeof(struct rx) * count, M_DEVBUF, M_NOWAIT);
	if (e100->rxs == NULL) {
		return errno;
	}
	memset(e100->rxs, 0x00, sizeof(struct rx) * count);
	return 0;
}

static int e100_rx_alloc_mbuf(e100_dev_t *e100, struct nw_work_thread *wtp, struct rx *rx)
{
	nic_stats_t *gstats = &e100->stats;

	if (!rx->m) {
		struct mbuf *m;
		dma_addr_t dma_addr;
		m = m_getcl_wtp(M_DONTWAIT, MT_DATA, M_PKTHDR, wtp);
		if (m == NULL) {
			gstats->rx_failed_allocs++;
			return ENOBUFS;
		}
		dma_addr = pool_phys(mtod(m, char*), m->m_ext.ext_page);
		rx->dma_addr = FROM_POINTER_TO_INT(NET_CACHELINE_ALIGN(dma_addr + RFD_ALIGNMENT_FUDGE), dma_addr_t);
		E100_DEBUG(
			log(LOG_DEBUG, "%s: %s() [%d] dma_addr 0x%08X (0x%08X) P 0x%llX V %p V0 %p", __devname__, __func__, RXID(rx), dma_addr, rx->dma_addr, m->m_ext.ext_page->pg_phys, mtod(m, char*), m->m_ext.ext_page->pg_page);
		)
		CACHE_INVAL(&e100->cachectl, mtod(m, char*), dma_addr, m->m_ext.ext_size);
		
		m->m_len = m->m_pkthdr.len = MCLBYTES;
		m_adj(m, (uintptr_t)NET_CACHELINE_ALIGN(mtod(m,char *) + RFD_ALIGNMENT_FUDGE) - mtod(m,uintptr_t) + sizeof(struct rfd));
		rx->m = m;
	} else {
		E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() [%d] dma_addr 0xXXXXXXXX (0x%08X) P 0x%llX V %p V0 %p", __devname__, __func__, RXID(rx), rx->dma_addr, rx->m->m_ext.ext_page->pg_phys, mtod(rx->m, char*), rx->m->m_ext.ext_page->pg_page);
		)
	}
	memcpy(RFD_FROM_M(rx->m), &e100->blank_rfd, sizeof(struct rfd));
	CACHE_FLUSH(&e100->cachectl, RFD_FROM_M(rx->m), rx->dma_addr, sizeof(struct rfd));

	/* Link the RFD to end of RFA by linking previous RFD to
	 * this one.  We are safe to touch the previous RFD because
	 * it is protected by the before last buffer's el bit being set */
	if (rx->prev->m) {
		struct rfd *prev_rfd = RFD_FROM_M(rx->prev->m);
		prev_rfd->link = cpu_to_le32(rx->dma_addr + e100->bmstr);
		CACHE_FLUSH(&e100->cachectl, prev_rfd, rx->prev->dma_addr, sizeof(struct rfd));
	}

	return 0;
}

int e100_rx_init_list(e100_dev_t *e100, struct nw_work_thread *wtp)
{
	struct rx *rx;
	unsigned int i, count = e100->params.rfds.count;
	struct rfd *before_last;

	e100->rx_to_use = e100->rx_to_clean = NULL;
	e100->ru_running = RU_UNINITIALIZED;

	for (rx = e100->rxs, i = 0; i < count; rx++, i++) {
		rx->next = (i + 1 < count) ? rx + 1 : e100->rxs;
		rx->prev = (i == 0) ? e100->rxs + count - 1 : rx - 1;
		if (e100_rx_alloc_mbuf(e100, wtp, rx)) {
			e100_rx_clean_list(e100, 0);
			return ENOMEM;
		}
	}
	/* Set the el-bit on the buffer that is before the last buffer.
	 * This lets us update the next pointer on the last buffer without
	 * worrying about hardware touching it.
	 * We set the size to 0 to prevent hardware from touching this buffer.
	 * When the hardware hits the before last buffer with el-bit and size
	 * of 0, it will RNR interrupt, the RU will go into the No Resources
	 * state.  It will not complete nor write to this buffer. */
	rx = e100->rxs->prev->prev;
	before_last = RFD_FROM_M(rx->m);
	before_last->command |= cpu_to_le16(cb_el);
	before_last->size = 0;
	CACHE_FLUSH(&e100->cachectl, before_last, rx->dma_addr, sizeof(struct rfd));

	e100->rx_to_use = e100->rx_to_clean = e100->rxs;
	e100->ru_running = RU_SUSPENDED;

	return 0;
}

static int e100_rx_indicate(e100_dev_t *e100, struct rx *rx,
	unsigned int *work_done, unsigned int work_to_do)
{
	struct ifnet *ifp = &e100->ecom.ec_if;
	struct mbuf *m = rx->m;
	struct rfd *rfd = RFD_FROM_M(m);
	nic_stats_t *gstats = &e100->stats;
	u16 rfd_status, actual_size;
	u16 fcs_pad = 0;
	uint8_t *dptr;

	if (unlikely(work_done && *work_done >= work_to_do))
		return EAGAIN;

	/* Need to sync before taking a peek at cb_complete bit */
	CACHE_INVAL(&e100->cachectl, RFD_FROM_M(rx->m), rx->dma_addr, sizeof(struct rfd));
	rfd_status = le16_to_cpu(rfd->status);

	dma_rmb(); /* read size after status bit */
	
	E100_DEBUG(
		log(LOG_DEBUG, "%s: [%d] stat 0x%04X cmd 0x%04X size 0x%02u cur 0x%08X link 0x%08X state %d", __devname__, RXID(rx), rfd_status, le16_to_cpu(rfd->command), le16_to_cpu(rfd->actual_size), rx->dma_addr, le32_to_cpu(rfd->link), e100->ru_running);
		dma_rmb(); /* read size after status bit */
	)

	/* If data isn't ready, nothing to indicate */
	if (unlikely(!(rfd_status & cb_complete))) {
		/* If the next buffer has the el bit, but we think the receiver
		 * is still running, check to see if it really stopped while
		 * we had interrupts off.
		 * This allows for a fast restart without re-enabling
		 * interrupts */
		if ((le16_to_cpu(rfd->command) & cb_el) &&
		    (RU_RUNNING == e100->ru_running)) {

			if (ioread8(&e100->csr->scb.status) & rus_no_res)
				e100->ru_running = RU_SUSPENDED;
			E100_DEBUG(
				log(LOG_DEBUG, "%s: !cb_complete cb_el state %d", __devname__, e100->ru_running);
			)
		} else if ( e100->extra_rnr ) {
			if (RU_RUNNING == e100->ru_running) {
				if (ioread8(&e100->csr->scb.status) & rus_no_res) {
					e100->ru_running = RU_SUSPENDED;
					E100_DEBUG(
						log(LOG_DEBUG, "%s: !cb_complete !cb_el no_res state %d", __devname__, e100->ru_running);
					)
				}
			}
		}
		CACHE_INVAL(&e100->cachectl, RFD_FROM_M(rx->m), rx->dma_addr, sizeof(struct rfd));
		return ENODATA;
	}

	/* Get actual data size */
// 	if (unlikely(dev->features & NETIF_F_RXFCS))
// 		fcs_pad = 4;
	actual_size = le16_to_cpu(rfd->actual_size) & 0x3FFF;
	if (unlikely(actual_size > RFD_BUF_LEN - sizeof(struct rfd)))
		actual_size = RFD_BUF_LEN - sizeof(struct rfd);

	/* If this buffer has the el bit, but we think the receiver
	 * is still running, check to see if it really stopped while
	 * we had interrupts off.
	 * This allows for a fast restart without re-enabling interrupts.
	 * This can happen when the RU sees the size change but also sees
	 * the el bit set. */
	if ((le16_to_cpu(rfd->command) & cb_el) &&
	    (RU_RUNNING == e100->ru_running)) {

	    if (ioread8(&e100->csr->scb.status) & rus_no_res)
		e100->ru_running = RU_SUSPENDED;
		E100_DEBUG(
		log(LOG_DEBUG, "%s: cb_complete cb_el state %d", __devname__, e100->ru_running);
		)
	}

	if (unlikely(!(rfd_status & cb_ok))) {
		/* Don't indicate if hardware indicates errors */
		m_freem(m);
	} else if (actual_size > MAX_ETHER_LEN + fcs_pad /*|| !actual_size *//*malformed packet*/) {
		/* Don't indicate oversized frames */
		e100->rx_over_length_errors++;
		m_freem(m);
	} else {
// process_mbuf:
		/* Put the actual info */
		m->m_pkthdr.rcvif = ifp;
		m->m_pkthdr.len = actual_size;
		m->m_len = m->m_pkthdr.len;

		/* stats */
		gstats->rxed_ok++;
		gstats->octets_rxed_ok += actual_size - fcs_pad;
		dptr = mtod(m, uint8_t *);
		if (dptr[0] & 1) {
			if (is_broadcast_mac(dptr))
				gstats->rxed_broadcast++;
			else
				gstats->rxed_multicast++;
		}
#if NBPFILTER > 0
		/* Pass this up to any BPF listeners. */
		if (ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif
		/* Send it up */
		ifp->if_ipackets++;
		(*ifp->if_input)(ifp, m);
		if (work_done)
			(*work_done)++;
	}
	rx->m = NULL;

	return 0;
}

void e100_rx_clean(e100_dev_t *e100, struct nw_work_thread *wtp, unsigned int *work_done, unsigned int work_to_do)
{
	struct rx *rx;
	int restart_required = 0, err = 0;
	struct rx *old_before_last_rx, *new_before_last_rx;
	struct rfd *old_before_last_rfd, *new_before_last_rfd;

	/* Indicate newly arrived packets */
	for (rx = e100->rx_to_clean; rx->m; rx = e100->rx_to_clean = rx->next) {
		err = e100_rx_indicate(e100, rx, work_done, work_to_do);
		/* Hit quota or no more to clean */
		if (EAGAIN == err || ENODATA == err)
			break;
	}
	E100_DEBUG2(
		if ( err == ENODATA ) {
			for (rx = e100->rx_to_clean; rx->m; rx = rx->next) {
				struct rfd *rfd = RFD_FROM_M(rx->m);
				u16 rfd_status;
				u16 rfd_command;

				/* Need to sync before taking a peek at cb_complete bit */
				CACHE_INVAL(&e100->cachectl, RFD_FROM_M(rx->m), rx->dma_addr, sizeof(struct rfd));
				rfd_status = le16_to_cpu(rfd->status);
				rfd_command = le16_to_cpu(rfd->command);
				if ( rfd_status || rfd_command ) {
					log(LOG_DEBUG, "%s: DBG: [%d] stat 0x%04X cmd 0x%04X state %d", __devname__, RXID(rx), rfd_status, rfd_command, e100->ru_running);
				}
				if ( rfd_command ) {
					break;
				}
			}
		}
	)


	/* On EAGAIN, hit quota so have more work to do, restart once
	 * cleanup is complete.
	 * Else, are we already rnr? then pay attention!!! this ensures that
	 * the state machine progression never allows a start with a
	 * partially cleaned list, avoiding a race between hardware
	 * and rx_to_clean when in NAPI mode */
	if (EAGAIN != err && RU_SUSPENDED == e100->ru_running)
		restart_required = 1;

	old_before_last_rx = e100->rx_to_use->prev->prev;
	if (old_before_last_rx->m) {
		old_before_last_rfd = RFD_FROM_M(old_before_last_rx->m);
	} else {
		old_before_last_rfd = NULL;
	}

	/* Alloc new mbufs to refill list */
	for (rx = e100->rx_to_use; !rx->m; rx = e100->rx_to_use = rx->next) {
		if (unlikely(e100_rx_alloc_mbuf(e100, wtp, rx)))
			break; /* Better luck next time (see watchdog) */
	}

	new_before_last_rx = e100->rx_to_use->prev->prev;
	if (new_before_last_rx != old_before_last_rx) {
		/* Set the el-bit on the buffer that is before the last buffer.
		 * This lets us update the next pointer on the last buffer
		 * without worrying about hardware touching it.
		 * We set the size to 0 to prevent hardware from touching this
		 * buffer.
		 * When the hardware hits the before last buffer with el-bit
		 * and size of 0, it will RNR interrupt, the RUS will go into
		 * the No Resources state.  It will not complete nor write to
		 * this buffer. */
		new_before_last_rfd = RFD_FROM_M(new_before_last_rx->m);
		E100_DEBUG(
			log(LOG_DEBUG, "%s: DBG: before_last_rx %d (0x%08X) => %d (0x%08X)", __devname__,
				RXID(old_before_last_rx), old_before_last_rx->dma_addr,
				RXID(new_before_last_rx), new_before_last_rx->dma_addr
			);
		)
		new_before_last_rfd->size = 0;
		new_before_last_rfd->command |= cpu_to_le16(cb_el);
		CACHE_FLUSH(&e100->cachectl, new_before_last_rfd, new_before_last_rx->dma_addr, sizeof(struct rfd));

		if ( old_before_last_rfd != NULL ) {
			/* Now that we have a new stopping point, we can clear the old
			* stopping point.  We must sync twice to get the proper
			* ordering on the hardware side of things. */
			old_before_last_rfd->command &= ~cpu_to_le16(cb_el);
			CACHE_FLUSH(&e100->cachectl, old_before_last_rfd, old_before_last_rx->dma_addr, sizeof(struct rfd));
			old_before_last_rfd->size = cpu_to_le16(MAX_ETHER_LEN);
			CACHE_FLUSH(&e100->cachectl, old_before_last_rfd, old_before_last_rx->dma_addr, sizeof(struct rfd));
		}
	}

	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() restart %d", __devname__, __func__, restart_required);
	)
	if (restart_required) {
		// ack the rnr?
		iowrite8(stat_ack_rnr, &e100->csr->scb.stat_ack);
		e100_start_receiver(e100, e100->rx_to_clean);
		if (work_done)
			(*work_done)++;
		if ( e100->cfg.verbose > 1 ) {
			log(LOG_DEBUG, "%s: receiver restart %d", __devname__, *work_done);
		}
	}
}

void e100_start_receiver(e100_dev_t *e100, struct rx *rx)
{
	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() start", __devname__, __func__);
	)
	if (!e100->rxs) return;
	if (RU_SUSPENDED != e100->ru_running) return;

	/* handle init time starts */
	if (!rx) rx = e100->rxs;

	/* (Re)start RU if suspended or idle and RFA is non-NULL */
	if (rx->m) {
		E100_DEBUG(
			log(LOG_DEBUG, "%s: %s() start with dma_addr P 0x%08X", __devname__, __func__, rx->dma_addr);
			while ( e100_exec_cmd(e100, ruc_abort, 0) ) {
				log(LOG_DEBUG, "%s: %s() ruc_abort!", __devname__, __func__);
			}
		)
		e100_exec_cmd(e100, ruc_start, rx->dma_addr);
		e100->ru_running = RU_RUNNING;
	}
	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() ru_running %d done", __devname__, __func__, e100->ru_running);
	)
}
