/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller transmit routines                                      *
 *                                                                           *
 *****************************************************************************/

#include "e100.h"


//
// this function is called only if the packet is ridiculously
// fragmented, and will fit into one cluster
//
static struct mbuf *e100_defrag(struct mbuf *m)
{
	struct mbuf *m2;

	if (m->m_pkthdr.len > MCLBYTES) {
		m_freem(m);
		return NULL;
	}

	MGET(m2, M_DONTWAIT, MT_DATA);
	if (m2 == NULL) {
		m_freem(m);
		return NULL;
	}

	M_COPY_PKTHDR(m2, m);

	MCLGET(m2, M_DONTWAIT);
	if ((m2->m_flags & M_EXT) == 0) {
		m_freem(m);
		m_freem(m2);
		return NULL;
	}

	m_copydata(m, 0, m->m_pkthdr.len, mtod(m2, caddr_t));
	m2->m_pkthdr.len = m2->m_len = m->m_pkthdr.len;

	m_freem(m);

	return m2;
}

static inline int mbuf_nfragments(struct mbuf *m0)
{
	int num_frag;

	struct mbuf *m = m0;

	for ( num_frag = 0; m != NULL; num_frag++ ) {
		m = m->m_next;
	}

	return num_frag;
}

static inline struct mbuf *ridiculously_fragmented_handling(e100_dev_t *e100, struct mbuf *m0)
{
	struct mbuf *m;
	int num_frag;

	num_frag = mbuf_nfragments(m0);
	if ( num_frag > E100_TX_FRAGMENTS ) {
		if ( e100->cfg.verbose > 1 )
			log(LOG_DEBUG, "%s: num_frag %d - ridiculously fragmented", __devname__, num_frag);

		if ( (m = e100_defrag(m0)) == NULL) {
			if (e100->cfg.verbose > 1)
				log(LOG_ERR, "%s: e100_defrag() failed", __devname__);
			goto done;
		} else if ( e100->cfg.verbose > 1 ) {
			log(LOG_ERR, "%s: warning: heavily fragmented normal packet transmitted: "
				"defrag worked: size %d, orignal num_frag %d", 
				__devname__, m->m_pkthdr.len, num_frag);
		}
		m0 = m;
	
		// must re-count mbuf fragments again
		num_frag = mbuf_nfragments(m0);
		if ( e100->cfg.verbose > 1 ) {
			log(LOG_ERR, "%s: warning: heavily fragmented normal packet transmitted: "
				"new num_frag %d", 
				__devname__, num_frag);
		}
		if ( num_frag > E100_TX_FRAGMENTS ) {
			if ( e100->cfg.verbose > 1 ) {
				log(LOG_ERR, "%s: dropped heavily fragmented huge packet: "
				"size %d, num_frag %d, free tx descr %d", 
				__devname__, m0->m_pkthdr.len, num_frag, e100->cbs_avail);
			}
			m_freem(m0);
			m0 = NULL;
			goto done;
		}
	}
done:
	return m0;
}

#if E100_DEBUG_LEVEL > 0
void dump_packet(e100_dev_t *e100, uint8_t *buf, size_t len, off64_t phys)
{
	int i = 0;
	size_t tail = len;

	log(LOG_ERR, "%s: buf %p len %zu phys 0x%"PRIX64, __devname__, buf, len, phys);

	for ( i = 0; i < len; ) {
		if ( tail >= 8 ) {
			log(LOG_ERR, "%s: %02X %02X %02X %02X %02X %02X %02X %02X", __devname__,
				buf[i + 0],
				buf[i + 1],
				buf[i + 2],
				buf[i + 3],
				buf[i + 4],
				buf[i + 5],
				buf[i + 6],
				buf[i + 7]
			);
			i += 8;
			tail -= 8;
		} else if ( tail >= 4 ) {
			log(LOG_ERR, "%s: %02X %02X %02X %02X", __devname__,
				buf[i + 0],
				buf[i + 1],
				buf[i + 2],
				buf[i + 3]
			);
			i += 4;
			tail -= 4;
		} else if ( tail >= 2 ) {
			log(LOG_ERR, "%s: %02X %02X", __devname__,
				buf[i + 0],
				buf[i + 1]
			);
			i += 2;
			tail -= 2;
		} else {
			log(LOG_ERR, "%s: %02X", __devname__,
				buf[i + 0]
			);
			i += 1;
			tail -= 1;
		}
	}
}
#endif

int e100_tx_clean(e100_dev_t *e100)
{
	struct cb *cb;
	struct ifnet *ifp = &e100->ecom.ec_if;
	struct mbuf *m;
	nic_stats_t *gstats = &e100->stats;
	int tx_cleaned = 0;
	int reaped = 0;
	uint8_t *dptr;

// 	InterruptLock(&e100->cb_lock);

	/* Clean CBs marked complete */
	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() cbs_avail %d", __devname__, __func__, e100->cbs_avail);
// 		for (cb = e100->cb_to_clean;
// 	    tx_cleaned < 16/*e100->params.cbs.count*/;
// 	    cb = cb->next) {
// 			log(LOG_DEBUG, "%s: CB[%d]->status = 0x%04X cmd 0x%04X", __devname__,
// 			     (int)(((void*)cb - (void*)e100->cbs)/sizeof(struct cb)),
// 			     cb->status, cb->command);
// 			tx_cleaned++;
// 		}
// 		tx_cleaned = 0;
	)
	for (cb = e100->cb_to_clean;
	    e100->params.cbs.reap >= reaped && (cb->status & cpu_to_le16(cb_complete));
	    cb = e100->cb_to_clean = cb->next) {
		dma_rmb(); /* read skb after status */
		if ( e100->cfg.verbose > 3 ) {
			log(LOG_DEBUG, "%s: cb[%d]->status = 0x%04X", __devname__,
			     (int)(((void*)cb - (void*)e100->cbs)/sizeof(struct cb)),
			     cb->status);
		}

		if (likely(cb->data != NULL)) {
			m = (struct mbuf *)cb->data;
// stat begin
			ifp->if_opackets++;
			gstats->txed_ok++;
			gstats->octets_txed_ok += m->m_pkthdr.len;
			dptr = mtod(m, uint8_t *);
			if (dptr[0] & 1) {
				if (is_broadcast_mac(dptr))
					gstats->txed_broadcast++;
				else
					gstats->txed_multicast++;
			}
// stat end

			m_freem(m);
			cb->data = NULL;
			tx_cleaned = 1;
		}
		cb->status = 0;
		e100->cbs_avail++;
		reaped++;
	}

// 	InterruptUnlock(&e100->cb_lock);

	return tx_cleaned;
}

E100_DEBUG2
(
static void log_dma_addr(e100_dev_t *e100, uint64_t paddr)
{
	log(LOG_DEBUG, "%s: %s() 0x%llX", __devname__, __func__, paddr);
}
)
static int e100_xmit_prepare(e100_dev_t *e100, struct cb *cb)
{
	int i;
	dma_addr_t dma_addr;
	struct mbuf *m0 = (struct mbuf *)cb->data;
	struct mbuf *m;
	cb->command = e100->tx_command;

	cb->command &= ~cpu_to_le16(cb_tx_nc);
	/* interrupt every 16 packets regardless of delay */

	if ((e100->cbs_avail & ~15) == e100->cbs_avail)
		cb->command |= cpu_to_le16(cb_i);

	i = 0;
	m = m0;
	while ( m != NULL ) {
		if ( !m->m_len ) {
			goto next_chunk;
		}
		dma_addr = (dma_addr_t)mbuf_phys(m);
		cb->u.tcb.tbd[i].buf_addr = cpu_to_le32(dma_addr + e100->bmstr);
		cb->u.tcb.tbd[i].size = cpu_to_le16(m->m_len);
		cb->u.tcb.tbd[i].eol = 0;
		E100_DEBUG2(
			log(LOG_DEBUG, "%s: %s() cb[%d]..tbd[%d].size %d(%d)", __devname__, __func__, (int)(((void*)cb - (void*)e100->cbs)/sizeof(struct cb)),  i, cb->u.tcb.tbd[i].size, (int)((uintptr_t)&cb->u.tcb.tbd[i].size - (uintptr_t)cb));
			log_dma_addr(e100, cb->dma_addr);
		)
		CACHE_FLUSH(&e100->cachectl, m->m_data, dma_addr, m->m_len);
		i++;
next_chunk:
		m = m->m_next;
	}
	cb->u.tcb.tbd_array = cb->dma_addr + offsetof(struct cb, u.tcb.tbd) + e100->bmstr;
	cb->u.tcb.tcb_byte_count = 0;
	cb->u.tcb.threshold = e100->tx_threshold;
	cb->u.tcb.tbd_count = i;
	E100_DEBUG(
		log(LOG_DEBUG, "%s: %s() tbd_array = 0x%08X i=%d", __devname__, __func__, cb->u.tcb.tbd_array, i);
	)

	return 0;
}

void e100_start(struct ifnet *ifp)
{
	e100_dev_t				*e100;
	struct mbuf				*m0;
	struct nw_work_thread	*wtp;
	int						ret;

	e100 = ifp->if_softc;
	wtp = WTP;

	if ( (ret = early_transmit(e100, wtp)) ) {
		if ( ret > 1 ) {
			IFQ_PURGE(&ifp->if_snd);
		}
		NW_SIGUNLOCK_P(&ifp->if_snd_ex, e100->iopkt, wtp);
		if ( e100->cfg.verbose > 1 )
			log(LOG_ERR, "%s: early transmit", __devname__);
		return;
	}

	e100->start_running = 1;
E100_DEBUG(
	log(LOG_ERR, "%s: start", __devname__);
)

	if (unlikely(!e100->cbs_avail)) {
		if ( e100->cfg.verbose > 8 )
			log(LOG_ERR, "%s: CB cleanup", __devname__);
		ifp->if_flags_tx |= IFF_OACTIVE;
		goto done; // not enough tx descriptors, try later
	}
	ifp->if_flags_tx |= IFF_OACTIVE;
	for (;;) {
		IFQ_POLL(&ifp->if_snd, m0);
		if (m0 == NULL) {
			if ( e100->cfg.verbose > 12 )
				log(LOG_ERR, "%s: IFQ_POLL 0", __devname__);
			break;
		}

		if (unlikely(!e100->cbs_avail)) {
			if ( e100->cfg.verbose > 1 )
				log(LOG_ERR, "%s: No space for CB", __devname__);
			ifp->if_flags_tx |= IFF_OACTIVE;
			e100->stats.tx_failed_allocs++;
// 			e100->tx_queue++;
			break;
		}
		/*
		 * Can look at m to see if you have the resources
		 * to transmit it.
		 */
		IFQ_DEQUEUE(&ifp->if_snd, m0);
#ifdef VARIANT_notx
		{
			m_freem(m0);
			e100->stats.tx_failed_allocs++;
			ifp->if_oerrors++;
			goto done;
		}
#endif
		if ( (m0 = ridiculously_fragmented_handling(e100, m0)) == NULL ) {
			e100->stats.tx_failed_allocs++;
			ifp->if_oerrors++;
			goto done;
		}

		if (e100->flags & ich_10h_workaround) {
			/* SW workaround for ICH[x] 10Mbps/half duplex Tx hang.
			Issue a NOP command followed by a 1us delay before
			issuing the Tx command. */
			if (e100_exec_cmd(e100, cuc_nop, 0))
				log(LOG_ERR, "%s: exec cuc_nop failed", __devname__);
			udelay(1);
		}

		ret = e100_exec_cb(e100, (void *)m0, e100_xmit_prepare);
		switch (ret) {
			case ENOSPC:
				if ( e100->cfg.verbose > 1 )
					log(LOG_DEBUG, "%s: No space for next CB", __devname__);
				break;
			case ENOMEM:
				e100->stats.tx_failed_allocs++;
				ifp->if_oerrors++;
				if ( e100->cfg.verbose > 1 )
					log(LOG_DEBUG, "%s: Out of Tx resources, returning mbuf", __devname__);
				goto done;
		}

#if NBPFILTER > 0
		/* Pass the packet to any BPF listeners. */
		if (ifp->if_bpf) {
			bpf_mtap(ifp->if_bpf, m0);
		}
#endif
	} // for

done:
	e100->start_running = 0;
	ifp->if_flags_tx &= ~IFF_OACTIVE;
E100_DEBUG (
	log(LOG_DEBUG, "%s: done", __devname__);
)
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, e100->iopkt, wtp);
}
